from e2cnn.gspaces import *
from e2cnn.nn import *
from e2cnn.group import *
from e2cnn.nn import init

import torch
import torch.nn as nn
import numpy as np


class ExpE2SFCNN(torch.nn.Module):
    
    def __init__(self, n_channels, n_classes,
                 layer_type: str = "regular",
                 N=None,
                 restrict: int = -1,
                 fix_param: bool = False,
                 fco: float = 0.8,
                 J: int = 0,
                 sigma: float = 0.6,
                 deltaorth: bool = False,
                 antialias: float = 0.,
                 sgsize: int = None,
                 flip: bool = True,
                 ):
        
        super(ExpE2SFCNN, self).__init__()
        
        if N is None:
            N = 16
        
        self.n_channels = n_channels
        self.n_classes = n_classes
        
        # build the group O(2) or D_N depending on the number N of rotations specified
        if N > 1:
            self.gspace = FlipRot2dOnR2(N)
        elif N <= 0:
            if restrict < -1:
                axis = 0.
                restrict = -1
            else:
                axis = np.pi / 2
            self.gspace = FlipRot2dOnR2(-1, maximum_frequency=-N, axis=axis)
            N = -1
        elif N == 1:
            self.gspace = Flip2dOnR2()
        else:
            raise ValueError(N)
        
        # if flips are not required, immediately restrict to the SO(2) or C_N subgroup
        if not flip:
            if N != 1:
                sg = (None, N)
            else:
                sg = 1
            self.gspace, _, _ = self.gspace.restrict(sg)

        if isinstance(fco, int) or isinstance(fco, float):
            fco *= np.pi
        self.frequencies_cutoff = fco
        self.sigma = sigma
        self.N = N
        self.restrict = restrict
        self.fix_param = fix_param
        self.J = J
        
        self.antialias = antialias
        
        # id of the subgroup if group restriction is applied through the network
        if sgsize is not None:
            self.sgid = sgsize
        elif N > 1 or N < 0:
            self.sgid = N
        else:
            self.sgid = 1
        
        if flip and N != 1:
            self.sgid = (None, self.sgid)

        eq_layers = []
        
        self.LAYER = 0
        
        r1 = FieldType(self.gspace, [self.gspace.trivial_repr] * n_channels)
        
        if layer_type == "regular":
            layer_builder = self.build_layer_regular
        elif layer_type == "quotient":
            layer_builder = self.build_layer_quotient
        elif layer_type == "gated_norm":
            layer_builder = self.build_layer_gated_normpool
        elif layer_type == "gated_norm_shared":
            layer_builder = self.build_layer_gated_normpool_shared
        elif layer_type == "gated_conv":
            layer_builder = self.build_layer_conv2triv
        elif layer_type == "gated_conv_shared":
            layer_builder = self.build_layer_gated_conv2triv_shared
        elif layer_type == "hnet_conv":
            layer_builder = self.build_layer_hnet_conv2triv
        elif layer_type == "hnet_norm":
            layer_builder = self.build_layer_hnet_normpool
        elif layer_type == "inducedhnet_conv":
            layer_builder = self.build_layer_inducedhnet_conv2triv
        elif layer_type == "inducedgated_norm":
            layer_builder = self.build_layer_inducedgated_normpool
        elif layer_type == "inducedgated_conv":
            layer_builder = self.build_layer_inducedgated_conv2triv
        elif layer_type == "sharednorm":
            layer_builder = self.build_layer_sharednorm
        elif layer_type == "sharednorm2":
            layer_builder = self.build_layer_sharednorm2
        elif layer_type == "vectorfield":
            layer_builder = self.build_layer_vectorfield
        elif layer_type == "regvector":
            layer_builder = self.build_layer_regvector
        elif layer_type == "scalarfield":
            layer_builder = self.build_layer_scalarfield
        elif layer_type == "trivial":
            layer_builder = self.build_layer_trivial
        elif layer_type == "squash":
            layer_builder = self.build_layer_squash
        elif layer_type == "realhnet":
            layer_builder = self.build_layer_realhnet
        elif layer_type == "realhnet2":
            layer_builder = self.build_layer_realhnet2
        elif layer_type == "debug":
            layer_builder = self.build_layer_debug
        else:
            raise ValueError(f"Error! Layer type {layer_type} not recognized!")
        
        # 28 px
        # Convolutional Layer 1
        
        self.LAYER += 1
        eq_layers += layer_builder(r1, 16, 7, 1)
        
        # 24 px
        # Convolutional Layer 2
        self.LAYER += 1
        eq_layers += layer_builder(eq_layers[-1].out_type, 24, 5, 2, 2)
        
        # 12 px
        # Convolutional Layer 3
        self.LAYER += 1
        eq_layers += layer_builder(eq_layers[-1].out_type, 32, 5, 2)
        
        # 12 px
        # Convolutional Layer 4
        self.LAYER += 1
        eq_layers += layer_builder(eq_layers[-1].out_type, 32, 5, 2, 2)
        
        # 6 px
        # Convolutional Layer 5
        self.LAYER += 1
        eq_layers += layer_builder(eq_layers[-1].out_type, 48, 5, 2)
        
        # 6 px
        # Convolutional Layer 6
        self.LAYER += 1
        eq_layers += layer_builder(eq_layers[-1].out_type, 64, 5, 0, None, True)
        
        # Adaptive Pooling
        mpl = PointwiseAdaptiveMaxPool(eq_layers[-1].out_type, 1)
        eq_layers.append(mpl)
        
        # 1 px
        
        # c = 64
        c = eq_layers[-1].out_type.size
        
        self.in_repr = eq_layers[0].in_type
        self.eq_layers = torch.nn.ModuleList(eq_layers)
        
        # Fully Connected
        
        self.fully_net = nn.Sequential(
            nn.Linear(c, 64),
            nn.BatchNorm1d(64),
            nn.ELU(inplace=True),
            
            nn.Linear(64, n_classes),
        )
        
        if deltaorth:
            for name, module in self.named_modules():
                if isinstance(module, R2Conv):
                    print(name)
                    init.deltaorthonormal_init(module.weights.data, module.basisexpansion)
        
        print("MODEL Architecture:")
        for i, (name, mod) in enumerate(self.named_modules()):
            params = sum([p.numel() for p in mod.parameters() if p.requires_grad])
            if isinstance(mod, EquivariantModule) and isinstance(mod.in_type, FieldType) and isinstance(mod.out_type,
                                                                                                        FieldType):
                print(f"\t{i: <3} - {name: <70} | {params: <8} | {mod.in_type.size: <4}- {mod.out_type.size: <4}")
            else:
                print(f"\t{i: <3} - {name: <70} | {params: <8} |")
        tot_param = sum([p.numel() for p in self.parameters() if p.requires_grad])
        print("Total number of parameters:", tot_param)
    
    def forward(self, input: torch.tensor):
        x = GeometricTensor(input, self.in_repr)
        
        for layer in self.eq_layers:
            x = layer(x)
        
        x = self.fully_net(x.tensor.reshape(x.tensor.shape[0], -1))
        
        return x
    
    def build_layer_regular(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                            invariant_map: bool = False):
        
        gc = r1.gspace
        
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # if self.fix_param and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            t = gc.fibergroup.order() / 16
            C = int(round(C / np.sqrt(t)))
        
        layers = []
        
        r2 = FieldType(gc, [gc.representations['regular']] * C)
        
        cl = R2Conv(r1, r2, s,
                    padding=padding,
                    frequencies_cutoff=self.frequencies_cutoff,
                    sigma=self.sigma,
                    maximum_offset=self.J,
                    )
        
        layers.append(cl)
        
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
        
        if invariant_map:
            pl = GroupPooling(layers[-1].out_type)
            layers.append(pl)
        
        bn = InnerBatchNorm(layers[-1].out_type)
        layers.append(bn)
        nnl = ELU(layers[-1].out_type)
        layers.append(nnl)
        
        if pooling is not None:
            if self.antialias > 0.:
                pl = PointwiseMaxPoolAntialiased(layers[-1].out_type, stride=pooling, sigma=self.antialias)
            elif self.antialias < 0.:
                pl = PointwiseAvgPoolAntialiased(layers[-1].out_type, stride=pooling, sigma=-self.antialias)
            else:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
        
        return layers

    def build_layer_quotient(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                             invariant_map: bool = False):
    
        gc = r1.gspace
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            t = gc.fibergroup.order() / 16
            C = C / np.sqrt(t)
    
        layers = []
    
        assert gc.fibergroup.order() > 0
        if isinstance(gc, FlipRot2dOnR2):
            n = int(gc.fibergroup.order() / 2)
            repr = [gc.regular_repr] * 5
            for i in [0, round(n / 4), round(n / 2)]:
                repr += [gc.quotient_repr((int(i), 1))] * 2
            if n % 2 == 0:
                repr += [gc.quotient_repr((None, 2))] * 2
            else:
                raise ValueError()
            repr += [gc.trivial_repr] * int(gc.fibergroup.order() / 4)
        elif isinstance(gc, Rot2dOnR2):
            n = gc.fibergroup.order()
            repr = [gc.regular_repr] * 5
            if n % 2 == 0:
                repr += [gc.quotient_repr(2)] * 2
            else:
                raise ValueError()
        
            if n % 4 == 0:
                repr += [gc.quotient_repr(4)] * 2
            elif n % 3 == 0:
                repr += [gc.quotient_repr(3)] * 2
            else:
                raise ValueError()
        
            repr += [gc.trivial_repr] * int(gc.fibergroup.order() / 4)
        else:
            repr = [gc.regular_repr]
    
        C /= sum([r.size for r in repr]) / gc.fibergroup.order()
    
        C = int(round(C))
    
        r2 = FieldType(gc, repr * C).sorted()
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if invariant_map:
            pl = GroupPooling(layers[-1].out_type)
            layers.append(pl)
    
        bn = InnerBatchNorm(layers[-1].out_type)
        layers.append(bn)
        nnl = ELU(layers[-1].out_type)
        layers.append(nnl)
    
        if pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
    
        return layers

    def build_layer_vectorfield(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                invariant_map: bool = False):
        
        # C_N equivariant model with vector-field non-linearity.
        
        assert isinstance(r1.gspace.fibergroup, CyclicGroup)
        
        gc = r1.gspace
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            # t = gc.fibergroup.order() / 16
            t = gc.fibergroup.order() / 16
            C /= np.sqrt(t)
    
        C *= np.sqrt(gc.fibergroup.order() / 2)
        C = int(round(C))
    
        layers = []
    
        r2 = FieldType(gc, [gc.representations['regular']] * C)
    
        cl = R2Conv(r1, r2, s,
                    padding=padding,
                    frequencies_cutoff=self.frequencies_cutoff,
                    sigma=self.sigma,
                    maximum_offset=self.J
                    )
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if invariant_map:
            pl = GroupPooling(layers[-1].out_type)
            layers.append(pl)
        
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
        
            nnl = ELU(layers[-1].out_type)
            layers.append(nnl)
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        else:
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
            nnl = VectorFieldNonLinearity(layers[-1].out_type)
            layers.append(nnl)
        
            if pooling is not None:
                pl = NormMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
    
        return layers

    def build_layer_scalarfield(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                invariant_map: bool = False):
        
        # equivariant model using group pooling at each layer
        
        assert r1.gspace.fibergroup.order() > 0
    
        gc = r1.gspace
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            # t = gc.fibergroup.order() / 16
            t = gc.fibergroup.order() / 16
            C /= np.sqrt(t)
    
        C *= np.sqrt(gc.fibergroup.order())
        C = int(round(C))
    
        layers = []
    
        r2 = FieldType(gc, [gc.representations['regular']] * C)
    
        cl = R2Conv(r1, r2, s,
                    padding=padding,
                    frequencies_cutoff=self.frequencies_cutoff,
                    sigma=self.sigma,
                    maximum_offset=self.J
                    )
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if invariant_map:
            pl = GroupPooling(layers[-1].out_type)
            layers.append(pl)
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
            nnl = ELU(layers[-1].out_type)
            layers.append(nnl)
        else:
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
            nnl = GroupPooling(layers[-1].out_type)
            layers.append(nnl)
            nnl = ELU(layers[-1].out_type)
            layers.append(nnl)
    
        if pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
    
        return layers

    def build_layer_regvector(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                              invariant_map: bool = False):
    
        RATIO = 0.5
    
        gc = r1.gspace
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            t = (RATIO * gc.fibergroup.order() + (1 - RATIO) * 2) / 16
            C /= np.sqrt(t)
    
        R = int(round(C * RATIO))
        V = int(round(C * (1 - RATIO)))
        C = V + R
    
        print(C, V, R)
    
        layers = []
    
        r2 = FieldType(gc, [gc.representations['regular']] * C)
    
        cl = R2Conv(r1, r2, s,
                    padding=padding,
                    frequencies_cutoff=self.frequencies_cutoff,
                    sigma=self.sigma,
                    maximum_offset=self.J
                    )
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if invariant_map:
            pl = GroupPooling(layers[-1].out_type)
            layers.append(pl)
        
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
        
            nnl = ELU(layers[-1].out_type, inplace=True)
            layers.append(nnl)
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        else:
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
        
            r = layers[-1].out_type
            assert R + V == len(r)
            labels = ["regular"] * R + ["vector"] * V
            r3 = r.group_by_labels(labels)
            regular = r3["regular"]
            vector = r3["vector"]
        
            modules = []
            modules += [(ELU(regular), "regular")]
            modules += [(VectorFieldNonLinearity(vector), "vector")]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(bn)
        
            if pooling is not None:
                r4 = layers[-1].out_type
                labels = ["vector" if r.irreducible else "regular" for r in r4]
                r4 = r4.group_by_labels(labels)
                regular = r4["regular"]
                vector = r4["vector"]
            
                modules = []
                modules += [(PointwiseMaxPool(regular, pooling), "regular")]
                modules += [(NormMaxPool(vector, pooling), "vector")]
                pl = MultipleModule(layers[-1].out_type, labels, modules)
                layers.append(pl)
    
        return layers

    def build_layer_hnet_normpool(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                  invariant_map: bool = False, function: str = "n_relu"):
        
        # NORM POOL in the end
        
        gc = r1.gspace
        
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
        irreps = list(irreps)
        
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [gc.trivial_repr] + irreps)
            r_out = FieldType(gc, [gc.trivial_repr] + irreps)
            
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
            
            t = tmp_cl.basisexpansion.dimension()
            
            t /= 16 * s ** 2 * 3 / 4
            
            C = int(round(C / np.sqrt(t)))
        
        elif invariant_map:
            # in order to preserve the same number of output channels
            size = sum(int(irr.size // irr.sum_of_squares_constituents) for irr in gc.fibergroup.irreps.values())
            C = int(round(C / size))
        
        layers = []
        
        irreps = list(irreps)
        
        trivials = FieldType(gc, [gc.trivial_repr] * C)
        if len(irreps) > 0:
            others = FieldType(gc, irreps * C).sorted()
            r2 = trivials + others
        else:
            others = []
            r2 = trivials
        
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
        
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
        
        r3 = layers[-1].out_type
        labels = ["trivial" if r.is_trivial() else "others" for r in r3]
        r3 = r3.group_by_labels(labels)
        trivials = r3["trivial"]
        others = r3["others"]
        
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
        
        modules = []
        modules += [(InnerBatchNorm(trivials), "trivial")]
        if len(irreps) > 0:
            modules += [(NormBatchNorm(others), "others")]
        bn = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(bn)
        
        labels = ["trivial" if r.is_trivial() else "others" for r in layers[-1].out_type]
        
        modules = []
        modules += [(ELU(trivials), "trivial")]
        if len(irreps) > 0:
            modules += [(NormNonLinearity(others, function=function), "others")]
        
        nnl = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(nnl)
        
        if invariant_map:
            modules = [
                (IdentityModule(trivials), "trivial"),
                (NormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
            
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            modules = []
            modules += [(PointwiseMaxPool(trivials, pooling), "trivial")]
            if len(irreps) > 0:
                modules += [(NormMaxPool(others, pooling), "others")]
            
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
        return layers

    def build_layer_squash(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                           invariant_map: bool = False):
        return self.build_layer_hnet_conv2triv(r1, C, s, padding, pooling, invariant_map, function="squash",
                                               norm_bias=False)

    def build_layer_hnet_conv2triv(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                   invariant_map: bool = False, function="n_relu", norm_bias=True):
    
        # conv2triv in the end
        # convolve to frequency-zero scalar field
        # if the group is O(2) (or D_N), convolve also to the sign-flip irreps and then takes the absolute value
        # of those fields
        
        gc = r1.gspace
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            irreps = []
            for n, irr in gc.fibergroup.irreps.items():
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
            r = FieldType(gc, irreps)
            tmp_cl = R2Conv(r, r, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        layers = []
    
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
    
        if len(irreps) > 0 and not invariant_map:
            trivials = FieldType(gc, [gc.trivial_repr] * C)
            others = FieldType(gc, irreps * C).sorted()
            r2 = trivials + others
        elif invariant_map:
            zero_freq = [irr for irr in gc.fibergroup.irreps.values() if irr.attributes['frequency'] == 0]
            irreps = [irr for irr in zero_freq if not irr.is_trivial()]
        
            if len(irreps) > 0:
                t = sum(irr.size for irr in zero_freq)
                t_i = sum(irr.size for irr in irreps)
                others = FieldType(gc, irreps * int(C * t_i / t)).sorted()
                trivials = FieldType(gc, [gc.trivial_repr] * int(C / t))
                r2 = trivials + others
            else:
                trivials = FieldType(gc, [gc.trivial_repr] * C)
                irreps = []
                others = None
                r2 = trivials
        else:
            trivials = FieldType(gc, [gc.trivial_repr] * C)
            irreps = []
            others = None
            r2 = trivials
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if len(irreps) > 0:
            r3 = layers[-1].out_type
            labels = ["trivial" if r.is_trivial() else "others" for r in r3]
            r3 = r3.group_by_labels(labels)
            trivials = r3["trivial"]
            others = r3["others"]
        else:
            trivials = layers[-1].out_type
            others = None
    
        assert all([r.is_trivial() for r in trivials])
    
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
    
        if len(irreps) > 0:
            modules = []
            modules += [(InnerBatchNorm(trivials), "trivial")]
            modules += [(NormBatchNorm(others), "others")]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(bn)
        else:
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
    
        if len(irreps) > 0:
            labels = ["trivial" if r.is_trivial() else "others" for r in layers[-1].out_type]
        
            modules = []
            modules += [(ELU(trivials), "trivial")]
            modules += [(NormNonLinearity(others, function=function, bias=norm_bias), "others")]
        
            nnl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(nnl)
    
        else:
            nnl = ELU(layers[-1].out_type)
            layers.append(nnl)
    
        if invariant_map and others is not None:
            modules = [
                (IdentityModule(trivials), "trivial"),
                (NormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
    
        elif len(irreps) == 0 and pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
        elif pooling is not None:
            modules = []
            modules += [(PointwiseMaxPool(trivials, pooling), "trivial")]
            modules += [(NormMaxPool(others, pooling), "others")]
        
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
    
        return layers

    def build_layer_sharednorm(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                               invariant_map: bool = False):
        
        # shared Norm-ReLU on irreps, ELU on trivial + Norm-Pooling"
        
        gc = r1.gspace
        
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
        irreps = list(irreps)
        
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [gc.trivial_repr] + irreps)
            r_out = FieldType(gc, [gc.trivial_repr] + irreps)
            
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
            
            t = tmp_cl.basisexpansion.dimension()
            
            t /= 16 * s ** 2 * 3 / 4
            
            C = int(round(C / np.sqrt(t)))
        
        elif invariant_map:
            
            # in order to preserve the same number of output channels
            if len(irreps) > 0:
                # there is a trivial field and a big field with all other frequencies
                size = 2
            else:
                # there is only a trivial field
                size = 1
            C = int(round(C / size))
        
        layers = []
        
        irreps_field = directsum(list(irreps), name="irreps")
        
        trivials = FieldType(gc, [gc.trivial_repr] * C)
        
        if len(irreps) > 0:
            others = FieldType(gc, [irreps_field] * C).sorted()
            r2 = trivials + others
        else:
            others = []
            r2 = trivials
            
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
        
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
        
        r3 = layers[-1].out_type
        labels = ["trivial" if r.is_trivial() else "others" for r in r3]
        r3 = r3.group_by_labels(labels)
        trivials = r3["trivial"]
        others = r3["others"]
        
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
        
        modules = []
        modules += [(InnerBatchNorm(trivials), "trivial")]
        if len(irreps) > 0:
            modules += [(NormBatchNorm(others), "others")]
        bn = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(bn)
        
        labels = ["trivial" if r.is_trivial() else "others" for r in layers[-1].out_type]
        
        modules = []
        modules += [(ELU(trivials), "trivial")]
        if len(irreps) > 0:
            modules += [(NormNonLinearity(others), "others")]
        
        nnl = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(nnl)
        
        if invariant_map:
            modules = [
                (IdentityModule(trivials), "trivial"),
                (NormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
            
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            modules = []
            modules += [(PointwiseMaxPool(trivials, pooling), "trivial")]
            if len(irreps) > 0:
                modules += [(NormMaxPool(others, pooling), "others")]
            
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
        return layers
    
    def build_layer_sharednorm2(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                invariant_map: bool = False):

        # shared Norm-ReLU on all irreps + Norm-Pooling
        
        gc = r1.gspace
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
        irreps = list(irreps)
        
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r = FieldType(gc, irreps)
            tmp_cl = R2Conv(r, r, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
            
            t = tmp_cl.basisexpansion.dimension()
            
            t /= 16 * s ** 2 / 2
            
            C = int(round(C / np.sqrt(t)))
        
        layers = []
        irreps = sorted(irreps, key=lambda i: i.size)
        irreps_field = directsum(irreps, name="irreps")
        
        r2 = FieldType(gc, [irreps_field] * C)
        
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
        
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
        
        bn = GNormBatchNorm(layers[-1].out_type)
        layers.append(bn)
        
        nnl = NormNonLinearity(layers[-1].out_type)
        layers.append(nnl)
        
        if invariant_map:
            pl = NormPool(layers[-1].out_type)
            layers.append(pl)
            
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            pl = NormMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
        
        return layers

    def build_layer_realhnet(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                             invariant_map: bool = False):
    
        gc = r1.gspace
    
        assert gc.fibergroup.order() < 0
        ################################################################################################################
        ############# NOTICE ###########################################################################################
        # The number of parameters is preserved as if Hnet3 is used.
        # This model, however, has a little less parameters as a more constrained kernel space is used
        ################################################################################################################
    
        if self.fix_param and self.LAYER > 1 and not invariant_map:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
        
            irreps = [irr for irr in gc.fibergroup.irreps.values()]
            tmp_repr = FieldType(gc, irreps)
        
            tmp_cl = R2Conv(tmp_repr, tmp_repr, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J,
                            )
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
            C = int(round(C / np.sqrt(t)))
    
        layers = []
    
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
    
        if len(irreps) > 0 and not invariant_map:
            trivials = FieldType(gc, [gc.trivial_repr] * C)
            others = FieldType(gc, irreps * C).sorted()
            r2 = trivials + others
        elif invariant_map:
            zero_freq = [irr for irr in gc.fibergroup.irreps.values() if irr.attributes['frequency'] == 0]
            irreps = [irr for irr in zero_freq if not irr.is_trivial()]
        
            if len(irreps) > 0:
                t = sum(irr.size for irr in zero_freq)
                t_i = sum(irr.size for irr in irreps)
                others = FieldType(gc, irreps * int(C * t_i / t)).sorted()
                trivials = FieldType(gc, [gc.trivial_repr] * int(C / t))
                r2 = trivials + others
            else:
                trivials = FieldType(gc, [gc.trivial_repr] * C)
                irreps = []
                others = None
                r2 = trivials
        else:
            trivials = FieldType(gc, [gc.trivial_repr] * C)
            irreps = []
            others = None
            r2 = trivials
    
        hnet_basis_filter = lambda d: d["s"] == 1 if "s" in d else True
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J,
                    basis_filter=hnet_basis_filter,
                    )
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if len(irreps) > 0:
            r3 = layers[-1].out_type
            labels = ["trivial" if r.is_trivial() else "others" for r in r3]
            r3 = r3.group_by_labels(labels)
            trivials = r3["trivial"]
            others = r3["others"]
        else:
            trivials = layers[-1].out_type
            others = None
    
        assert all([r.is_trivial() for r in trivials])
    
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
    
        if len(irreps) > 0:
            modules = []
            modules += [(InnerBatchNorm(trivials), "trivial")]
            modules += [(NormBatchNorm(others), "others")]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(bn)
        else:
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
    
        if len(irreps) > 0:
            labels = ["trivial" if r.is_trivial() else "others" for r in layers[-1].out_type]
        
            modules = []
            modules += [(ELU(trivials), "trivial")]
            modules += [(NormNonLinearity(others), "others")]
        
            nnl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(nnl)
    
        else:
            nnl = ELU(layers[-1].out_type)
            layers.append(nnl)
    
        if invariant_map and others is not None:
            modules = [
                (IdentityModule(trivials), "trivial"),
                (NormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
    
        elif len(irreps) == 0 and pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
        elif pooling is not None:
            modules = []
            modules += [(PointwiseMaxPool(trivials, pooling), "trivial")]
            modules += [(NormMaxPool(others, pooling), "others")]
        
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
    
        return layers

    def build_layer_realhnet2(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                              invariant_map: bool = False):
    
        gc = r1.gspace
    
        assert gc.fibergroup.order() < 0
    
        hnet_basis_filter = lambda d: d["s"] == 1 if "s" in d else True
    
        ################################################################################################################
        ############# NOTICE ###########################################################################################
        # WRT "realhnet", here we scale up the model to match the number of parameters.
        ################################################################################################################
    
        if self.fix_param and self.LAYER > 1 and not invariant_map:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
        
            irreps = [irr for irr in gc.fibergroup.irreps.values()]
            tmp_repr = FieldType(gc, irreps)
        
            tmp_cl = R2Conv(tmp_repr, tmp_repr, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J,
                            basis_filter=hnet_basis_filter,
                            )
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        layers = []
    
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
    
        if len(irreps) > 0 and not invariant_map:
            trivials = FieldType(gc, [gc.trivial_repr] * C)
            others = FieldType(gc, irreps * C).sorted()
            r2 = trivials + others
        elif invariant_map:
            zero_freq = [irr for irr in gc.fibergroup.irreps.values() if irr.attributes['frequency'] == 0]
            irreps = [irr for irr in zero_freq if not irr.is_trivial()]
        
            if len(irreps) > 0:
                t = sum(irr.size for irr in zero_freq)
                t_i = sum(irr.size for irr in irreps)
                others = FieldType(gc, irreps * int(C * t_i / t)).sorted()
                trivials = FieldType(gc, [gc.trivial_repr] * int(C / t))
                r2 = trivials + others
            else:
                trivials = FieldType(gc, [gc.trivial_repr] * C)
                irreps = []
                others = None
                r2 = trivials
        else:
            trivials = FieldType(gc, [gc.trivial_repr] * C)
            irreps = []
            others = None
            r2 = trivials
    
        print(r2.size)
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J,
                    basis_filter=hnet_basis_filter,
                    )
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if len(irreps) > 0:
            r3 = layers[-1].out_type
            labels = ["trivial" if r.is_trivial() else "others" for r in r3]
            r3 = r3.group_by_labels(labels)
            trivials = r3["trivial"]
            others = r3["others"]
        else:
            trivials = layers[-1].out_type
            others = None
    
        assert all([r.is_trivial() for r in trivials])
    
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
    
        if len(irreps) > 0:
            modules = []
            modules += [(InnerBatchNorm(trivials), "trivial")]
            modules += [(NormBatchNorm(others), "others")]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(bn)
        else:
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
    
        if len(irreps) > 0:
            labels = ["trivial" if r.is_trivial() else "others" for r in layers[-1].out_type]
        
            modules = []
            modules += [(ELU(trivials), "trivial")]
            modules += [(NormNonLinearity(others), "others")]
        
            nnl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(nnl)
    
        else:
            nnl = ELU(layers[-1].out_type)
            layers.append(nnl)
    
        if invariant_map and others is not None:
            modules = [
                (IdentityModule(trivials), "trivial"),
                (NormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
    
        elif len(irreps) == 0 and pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
        elif pooling is not None:
            modules = []
            modules += [(PointwiseMaxPool(trivials, pooling), "trivial")]
            modules += [(NormMaxPool(others, pooling), "others")]
        
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
    
        return layers

    def build_layer_gated_normpool(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                   invariant_map: bool = False):
    
        ############################################################
        # 1 gate per irrep (except trivial); ELU on trivial irreps
        ############################################################
    
        gc = r1.gspace
    
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
        irreps = list(irreps)
    
        # number of fields requiring a gate
        I = len(irreps)
    
        # size of all irreps (including trivial)
        S = FieldType(gc, irreps).size + 1
        M = S + I
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [gc.trivial_repr] * (I + 1) + irreps)
            r_out = FieldType(gc, [gc.trivial_repr] + irreps)
        
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        elif invariant_map:
            # in order to preserve the same number of output channels
            size = sum(int(irr.size // irr.sum_of_squares_constituents) for irr in gc.fibergroup.irreps.values())
            C = int(round(C / size))
    
        layers = []
    
        # total channels =   C * S    +   1 * C * I     = C * M
        # i.e.:            (fields) + (gates for non-trivial fields)
    
        trivials = FieldType(gc, [gc.trivial_repr] * C)
        gates = FieldType(gc, [gc.trivial_repr] * C * I)
        gated = FieldType(gc, irreps * C).sorted()
        gate = gates + gated
    
        r2 = trivials + gate
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        labels = ["trivial"] * (len(trivials) + len(gates)) + ["gated"] * len(gated)
    
        modules = [
            (InnerBatchNorm(trivials + gates), "trivial"),
            (NormBatchNorm(gated), "gated")
        ]
        bn = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(bn)
    
        labels = ["trivial"] * len(trivials) + ["gate"] * len(gate)
        modules = [
            (ELU(trivials), "trivial"),
            (GatedNonLinearity1(gate), "gate")
        ]
        nnl = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(nnl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        r3 = layers[-1].out_type
        labels = ["trivial" if r.is_trivial() else "others" for r in r3]
        r3 = r3.group_by_labels(labels)
        trivials = r3["trivial"]
        others = r3["others"]
    
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
    
        if invariant_map:
            modules = [
                (IdentityModule(trivials), "trivial"),
                (NormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            modules = [
                (PointwiseMaxPool(trivials, pooling), "trivial"),
                (NormMaxPool(others, pooling), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
    
        return layers

    def build_layer_gated_normpool_shared(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                          invariant_map: bool = False):
    
        ###################################################################################
        # 1 gate per field containing all irreps except trivial; ELU on trivial irreps
        ###################################################################################
    
        gc = r1.gspace
    
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
        irreps = list(irreps)
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [gc.trivial_repr] * 2 + irreps)
            r_out = FieldType(gc, [gc.trivial_repr] + irreps)
        
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        elif invariant_map:
            # in order to preserve the same number of output channels
            size = sum(int(irr.size // irr.sum_of_squares_constituents) for irr in gc.fibergroup.irreps.values())
            C = int(round(C / size))
    
        layers = []
    
        irreps_field = directsum(list(irreps), name="irreps")
    
        trivials = FieldType(gc, [gc.trivial_repr] * C)
        gates = FieldType(gc, [gc.trivial_repr] * C)
        gated = FieldType(gc, [irreps_field] * C).sorted()
        gate = gates + gated
    
        r2 = trivials + gate
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        labels = ["trivial"] * (len(trivials) + len(gates)) + ["gated"] * len(gated)
    
        modules = [
            (InnerBatchNorm(trivials + gates), "trivial"),
            (NormBatchNorm(gated), "gated")
        ]
        bn = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(bn)
    
        labels = ["trivial"] * len(trivials) + ["gate"] * len(gate)
        modules = [
            (ELU(trivials), "trivial"),
            (GatedNonLinearity1(gate), "gate")
        ]
        nnl = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(nnl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        r3 = layers[-1].out_type
        labels = ["trivial" if r.is_trivial() else "others" for r in r3]
        r3 = r3.group_by_labels(labels)
        trivials = r3["trivial"]
        others = r3["others"]
    
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
    
        if invariant_map:
            modules = [
                (IdentityModule(trivials), "trivial"),
                (NormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            modules = [
                (PointwiseMaxPool(trivials, pooling), "trivial"),
                (NormMaxPool(others, pooling), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
    
        return layers

    def build_layer_conv2triv(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                              invariant_map: bool = False):
    
        ############################################################
        # 1 gate per irrep (except trivial); ELU on trivial irreps
        # conv 2 triv in the end
        ############################################################
    
        gc = r1.gspace
    
        irreps = []
        for n, irr in gc.fibergroup.irreps.items():
            if n != gc.trivial_repr.name:
                if not invariant_map or irr.attributes['frequency'] == 0:
                    irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
        irreps = list(irreps)
    
        # number of fields requiring a gate
        I = len(irreps)
    
        # size of all irreps (including trivial)
        S = sum(r.size for r in irreps) + 1
        M = S + I
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [gc.trivial_repr] * (I + 1) + irreps)
            r_out = FieldType(gc, [gc.trivial_repr] + irreps)
        
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        elif invariant_map:
            # in order to preserve the same number of output channels
            size = sum(int(irr.size // irr.sum_of_squares_constituents) for irr in gc.fibergroup.irreps.values() if
                       irr.attributes["frequency"] == 0)
            C = int(round(C / size))
    
        layers = []
    
        # total channels =   C * S    +   1 * C * I     = C * M
        # i.e.:            (fields) + (gates for non-trivial fields)
    
        trivials = FieldType(gc, [gc.trivial_repr] * C)
        if len(irreps) > 0:
            gates = FieldType(gc, [gc.trivial_repr] * C * I)
            gated = FieldType(gc, irreps * C).sorted()
            gate = gates + gated
        
            r2 = trivials + gate
        else:
            gates = gated = gate = []
            r2 = trivials
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        if len(irreps) > 0:
            labels = ["trivial"] * (len(trivials) + len(gates)) + ["gated"] * len(gated)
        
            modules = [
                (InnerBatchNorm(trivials + gates), "trivial"),
                (NormBatchNorm(gated), "gated")
            ]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
        else:
            bn = InnerBatchNorm(trivials)
        layers.append(bn)
    
        if len(irreps) > 0:
            labels = ["trivial"] * len(trivials) + ["gate"] * len(gate)
            modules = [
                (ELU(trivials), "trivial"),
                (GatedNonLinearity1(gate), "gate")
            ]
            nnl = MultipleModule(layers[-1].out_type, labels, modules)
        else:
            nnl = ELU(trivials)
        layers.append(nnl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        r3 = layers[-1].out_type
        labels = ["trivial" if r.is_trivial() else "others" for r in r3]
        r3 = r3.group_by_labels(labels)
        trivials = r3["trivial"]
        if "others" in r3:
            others = r3["others"]
        else:
            others = []
    
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
    
        if invariant_map:
            if len(others) > 0:
                modules = [
                    (IdentityModule(trivials), "trivial"),
                    (NormPool(others), "others")
                ]
                pl = MultipleModule(layers[-1].out_type, labels, modules)
                layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            if len(others) > 0:
                modules = [
                    (PointwiseMaxPool(trivials, pooling), "trivial"),
                    (NormMaxPool(others, pooling), "others")
                ]
                pl = MultipleModule(layers[-1].out_type, labels, modules)
            else:
                pl = PointwiseMaxPool(trivials, pooling)
            layers.append(pl)
    
        return layers

    def build_layer_gated_conv2triv_shared(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                           invariant_map: bool = False):
    
        ###################################################################################
        # 1 gate per field containing all irreps except trivial; ELU on trivial irreps
        # conv 2 triv in the end
        ###################################################################################
    
        gc = r1.gspace
        assert not isinstance(gc, FlipRot2dOnR2)
    
        irreps = []
        if not invariant_map:
            for n, irr in gc.fibergroup.irreps.items():
                if n != gc.trivial_repr.name:
                    irreps += [irr] * int(irr.size // irr.sum_of_squares_constituents)
        irreps = list(irreps)
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [gc.trivial_repr] * 2 + irreps)
            r_out = FieldType(gc, [gc.trivial_repr] + irreps)
        
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        layers = []
    
        trivials = FieldType(gc, [gc.trivial_repr] * C)
    
        if len(irreps) > 0:
            irreps_field = directsum(list(irreps), name="irreps")
        
            gates = FieldType(gc, [gc.trivial_repr] * C)
            gated = FieldType(gc, [irreps_field] * C).sorted()
            gate = gates + gated
        
            r2 = trivials + gate
        else:
            r2 = trivials
            gate = gated = gates = []
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        if len(irreps) > 0:
            labels = ["trivial"] * (len(trivials) + len(gates)) + ["gated"] * len(gated)
        
            modules = [
                (InnerBatchNorm(trivials + gates), "trivial"),
                (NormBatchNorm(gated), "gated")
            ]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
        else:
            bn = InnerBatchNorm(trivials)
        layers.append(bn)
    
        if len(irreps) > 0:
            labels = ["trivial"] * len(trivials) + ["gate"] * len(gate)
            modules = [
                (ELU(trivials), "trivial"),
                (GatedNonLinearity1(gate), "gate")
            ]
            nnl = MultipleModule(layers[-1].out_type, labels, modules)
        else:
            nnl = ELU(trivials)
        layers.append(nnl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        r3 = layers[-1].out_type
        labels = ["trivial" if r.is_trivial() else "others" for r in r3]
        r3 = r3.group_by_labels(labels)
        trivials = r3["trivial"]
        if "others" in r3:
            others = r3["others"]
        else:
            others = []
    
        for r in trivials:
            r.supported_nonlinearities.add("pointwise")
    
        if invariant_map:
            if len(others) > 0:
                modules = [
                    (IdentityModule(trivials), "trivial"),
                    (NormPool(others), "others")
                ]
                pl = MultipleModule(layers[-1].out_type, labels, modules)
                layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            if len(others) > 0:
                modules = [
                    (PointwiseMaxPool(trivials, pooling), "trivial"),
                    (NormMaxPool(others, pooling), "others")
                ]
                pl = MultipleModule(layers[-1].out_type, labels, modules)
            else:
                pl = PointwiseMaxPool(trivials, pooling)
            layers.append(pl)
    
        return layers

    def build_layer_trivial(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                            invariant_map: bool = False):
    
        gc = r1.gspace
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant
            # C *= np.sqrt(16 * 2*s)
            C *= np.sqrt(16 * s)
    
        C = int(round(C))
    
        layers = []
    
        r2 = FieldType(gc, [gc.trivial_repr] * C)
    
        cl = R2Conv(r1, r2, s,
                    padding=padding,
                    frequencies_cutoff=self.frequencies_cutoff,
                    sigma=self.sigma,
                    maximum_offset=self.J
                    )
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        bn = InnerBatchNorm(layers[-1].out_type)
        layers.append(bn)
        nnl = ELU(layers[-1].out_type, inplace=True)
        layers.append(nnl)
    
        if pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
    
        return layers

    def build_layer_inducedhnet_conv2triv(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                          invariant_map: bool = False):
        
        # individual Norm-ReLU with bias shared within induced field,
        # conv to induced trivial (+gpool).
        # Fix total number of params
        
        gc = r1.gspace
        assert isinstance(gc, FlipRot2dOnR2)
        
        if self.fix_param and self.LAYER > 1 and not invariant_map:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            F = max([r.attributes["frequency"] for r in gc.fibergroup.irreps.values()])
            M = 3 * s / 2
            B = F * (2 * M + 1) - M * min(M, F)
            
            t = 4 * B / 6 + 2 * min(F, M) / 6 + 1
            t *= s / 2
            t *= sum([r.size ** 2 // r.sum_of_squares_constituents for r in gc.fibergroup.irreps.values()])
            t /= 2 * F + 1
            
            t /= 16 * s ** 2 / 2
            C = (C / np.sqrt(t))
        
        C = int(round(C))
        
        subgroup_id = (None, gc.fibergroup.rotation_order)
        sg, _, _ = gc.restrict(subgroup_id)
        
        layers = []
        
        irreps = []
        for n, irr in sg.fibergroup.irreps.items():
            if n != sg.trivial_repr.name:
                irreps += [gc.induced_repr(subgroup_id, irr)]
        
        trivials = FieldType(gc, [gc.induced_repr(subgroup_id, sg.trivial_repr)] * C)
        
        if len(irreps) > 0 and not invariant_map:
            others = FieldType(gc, irreps * C).sorted()
            r2 = trivials + others
        else:
            irreps = []
            others = None
            r2 = trivials
        
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
        
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
        
        def only_zero_freq(repr: Representation):
            for irr in repr.irreps:
                if repr.group.irreps[irr].attributes['frequency'] != 0:
                    return False
            return True
        
        if len(irreps) > 0:
            r3 = layers[-1].out_type
            labels = ["trivial" if only_zero_freq(r) else "others" for r in r3]
            r3 = r3.group_by_labels(labels)
            trivials = r3["trivial"]
            others = r3["others"]
        else:
            trivials = layers[-1].out_type
            others = None
        
        assert all([only_zero_freq(r) for r in trivials])
        
        if len(irreps) > 0:
            modules = []
            modules += [(InnerBatchNorm(trivials), "trivial")]
            modules += [(InducedNormBatchNorm(others), "others")]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(bn)
        else:
            bn = InnerBatchNorm(layers[-1].out_type)
            layers.append(bn)
        
        if len(irreps) > 0:
            labels = ["trivial" if only_zero_freq(r) else "others" for r in layers[-1].out_type]
            
            modules = []
            modules += [(ELU(trivials), "trivial")]
            modules += [(InducedNormNonLinearity(others), "others")]
            
            nnl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(nnl)
        
        else:
            nnl = ELU(layers[-1].out_type)
            layers.append(nnl)
        
        if invariant_map:
            opool = GroupPooling(layers[-1].out_type)
            layers.append(opool)
            
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        
        elif len(irreps) == 0 and pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
        elif pooling is not None:
            modules = []
            modules += [(PointwiseMaxPool(trivials, pooling), "trivial")]
            modules += [(NormMaxPool(others, pooling), "others")]
            
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
        return layers

    def build_layer_inducedgated_normpool(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None, invariant_map: bool = False):
        gc = r1.gspace
        assert isinstance(gc, FlipRot2dOnR2)

        ############################################################
        # 1 gate per irrep (except trivial); ELU on trivial irreps
        ############################################################
    
        gc = r1.gspace
        
        subgroup_id = (None, gc.fibergroup.rotation_order)
        sg, _, _ = gc.restrict(subgroup_id)

        layers = []

        irreps = []
        for n, irr in sg.fibergroup.irreps.items():
            if not irr.is_trivial():
                irreps += [gc.induced_repr(subgroup_id, irr)]

        irreps = list(irreps)

        induced_trivial = gc.induced_repr(subgroup_id, sg.trivial_repr)

        # number of fields requiring a gate
        I = len(irreps)
    
        # size of all irreps (including trivial)
        S = FieldType(gc, irreps).size + induced_trivial.size
    
        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [induced_trivial] * (I + 1) + irreps)
            r_out = FieldType(gc, [induced_trivial] + irreps)
        
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        elif invariant_map:
            # in order to preserve the same number of output channels
            size = I + 1
            C = int(round(C / size))
    
        layers = []
    
        # total channels =   C * S    +   1 * C * I     = C * M
        # i.e.:            (fields) + (gates for non-trivial fields)
    
        trivials = FieldType(gc, [induced_trivial] * C)
        gates = FieldType(gc, [induced_trivial] * C * I)
        gated = FieldType(gc, irreps * C).sorted()
        gate = gates + gated
    
        r2 = trivials + gate
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        labels = ["trivial"] * (len(trivials) + len(gates)) + ["gated"] * len(gated)
    
        modules = [
            (InnerBatchNorm(trivials + gates), "trivial"),
            (InducedNormBatchNorm(gated), "gated")
        ]
        bn = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(bn)
    
        labels = ["trivial"] * len(trivials) + ["gate"] * len(gate)
        modules = [
            (ELU(trivials), "trivial"),
            (InducedGatedNonLinearity1(gate), "gate")
        ]
        nnl = MultipleModule(layers[-1].out_type, labels, modules)
        layers.append(nnl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))

        def only_zero_freq(repr: Representation):
            for irr in repr.irreps:
                if repr.group.irreps[irr].attributes['frequency'] != 0:
                    return False
            return True

        if len(irreps) > 0:
            r3 = layers[-1].out_type
            labels = ["trivial" if only_zero_freq(r) else "others" for r in r3]
            r3 = r3.group_by_labels(labels)
            trivials = r3["trivial"]
            others = r3["others"]
        else:
            trivials = layers[-1].out_type
            others = None

        assert all([only_zero_freq(r) for r in trivials])

        if invariant_map:
            modules = [
                (GroupPooling(trivials), "trivial"),
                (InducedNormPool(others), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
        elif pooling is not None:
            modules = [
                (PointwiseMaxPool(trivials, pooling), "trivial"),
                (NormMaxPool(others, pooling), "others")
            ]
            pl = MultipleModule(layers[-1].out_type, labels, modules)
            layers.append(pl)
    
        return layers

    def build_layer_inducedgated_conv2triv(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                                           invariant_map: bool = False):
    
        ############################################################
        # 1 gate per irrep (except trivial); ELU on trivial irreps
        # conv 2 triv in the end
        ############################################################
    
        gc = r1.gspace
        assert isinstance(gc, FlipRot2dOnR2)
        
        subgroup_id = (None, gc.fibergroup.rotation_order)
        sg, _, _ = gc.restrict(subgroup_id)

        layers = []

        irreps = []
        if not invariant_map:
            for n, irr in sg.fibergroup.irreps.items():
                if not irr.is_trivial():
                    irreps += [gc.induced_repr(subgroup_id, irr)]

        induced_trivial = gc.induced_repr(subgroup_id, sg.trivial_repr)

        # number of fields requiring a gate
        I = len(irreps)

        # size of all irreps (including trivial)
        S = sum(r.size for r in irreps) + induced_trivial.size

        if self.fix_param and not invariant_map and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            r_in = FieldType(gc, [induced_trivial] * (I + 1) + irreps)
            r_out = FieldType(gc, [induced_trivial] + irreps)
        
            tmp_cl = R2Conv(r_in, r_out, s,
                            frequencies_cutoff=self.frequencies_cutoff,
                            padding=padding,
                            sigma=self.sigma,
                            maximum_offset=self.J)
        
            t = tmp_cl.basisexpansion.dimension()
        
            t /= 16 * s ** 2 * 3 / 4
        
            C = int(round(C / np.sqrt(t)))
    
        layers = []
    
        # total channels =   C * S    +   1 * C * I     = C * M
        # i.e.:            (fields) + (gates for non-trivial fields)
    
        trivials = FieldType(gc, [induced_trivial] * C)
        if len(irreps) > 0:
            gates = FieldType(gc, [induced_trivial] * C * I)
            gated = FieldType(gc, irreps * C).sorted()
            gate = gates + gated
        
            r2 = trivials + gate
        else:
            gates = gated = gate = []
            r2 = trivials
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        if len(irreps) > 0:
            labels = ["trivial"] * (len(trivials) + len(gates)) + ["gated"] * len(gated)
        
            modules = [
                (InnerBatchNorm(trivials + gates), "trivial"),
                (InducedNormBatchNorm(gated), "gated")
            ]
            bn = MultipleModule(layers[-1].out_type, labels, modules)
        else:
            bn = InnerBatchNorm(trivials)
            
        layers.append(bn)
    
        if len(irreps) > 0:
            labels = ["trivial"] * len(trivials) + ["gate"] * len(gate)
            modules = [
                (ELU(trivials), "trivial"),
                (InducedGatedNonLinearity1(gate), "gate")
            ]
            nnl = MultipleModule(layers[-1].out_type, labels, modules)
        else:
            nnl = ELU(trivials)
        layers.append(nnl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
            
        def only_zero_freq(repr: Representation):
            for irr in repr.irreps:
                if repr.group.irreps[irr].attributes['frequency'] != 0:
                    return False
            return True

        if len(irreps) > 0:
            r3 = layers[-1].out_type
            labels = ["trivial" if only_zero_freq(r) else "others" for r in r3]
            r3 = r3.group_by_labels(labels)
            trivials = r3["trivial"]
            others = r3["others"]
        else:
            trivials = layers[-1].out_type
            others = []

        assert all([only_zero_freq(r) for r in trivials])

        if invariant_map:
            
            pl = GroupPooling(trivials)
            layers.append(pl)
        
            if pooling is not None:
                pl = PointwiseMaxPool(layers[-1].out_type, pooling)
                layers.append(pl)
                
        elif pooling is not None:
            if len(others) > 0:
                modules = [
                    (PointwiseMaxPool(trivials, pooling), "trivial"),
                    (NormMaxPool(others, pooling), "others")
                ]
                pl = MultipleModule(layers[-1].out_type, labels, modules)
            else:
                pl = PointwiseMaxPool(trivials, pooling)
            layers.append(pl)
    
        return layers
    
    def build_layer_debug(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                          invariant_map: bool = False):
        
        gc = r1.gspace
        
        C /= 4
        
        if self.fix_param and not invariant_map and self.LAYER > 1:
            t = gc.fibergroup.order() / 16
            C = C / np.sqrt(t)
        
        C = int(round(C))
        
        layers = []
        
        r2 = FieldType(gc, [gc.representations['regular']] * C)
        
        cl = R2Conv(r1, r2, s,
                    padding=padding,
                    frequencies_cutoff=self.frequencies_cutoff,
                    sigma=self.sigma,
                    maximum_offset=self.J
                    )
        layers.append(cl)
        
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
        
        if invariant_map:
            pl = GroupPooling(layers[-1].out_type)
            layers.append(pl)
        
        nnl = ReLU(layers[-1].out_type, inplace=False)
        layers.append(nnl)
        
        if pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
        
        return layers




# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb
import numpy as np
import matplotlib.pyplot as pl


def make_colorwheel():
    '''
    Generates a color wheel for optical flow visualization as presented in:
        Baker et al. "A Database and Evaluation Methodology for Optical Flow" (ICCV, 2007)
        URL: http://vision.middlebury.edu/flow/flowEval-iccv07.pdf
    According to the C++ source code of Daniel Scharstein
    According to the Matlab source code of Deqing Sun
    
    Copied from https://github.com/tomrunia/OpticalFlow_Visualization/blob/master/flow_vis.py
    Copyright (c) 2018 Tom Runia
    '''

    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR
    colorwheel = np.zeros((ncols, 3))
    col = 0

    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.floor(255*np.arange(0,RY)/RY)
    col = col+RY
    # YG
    colorwheel[col:col+YG, 0] = 255 - np.floor(255*np.arange(0,YG)/YG)
    colorwheel[col:col+YG, 1] = 255
    col = col+YG
    # GC
    colorwheel[col:col+GC, 1] = 255
    colorwheel[col:col+GC, 2] = np.floor(255*np.arange(0,GC)/GC)
    col = col+GC
    # CB
    colorwheel[col:col+CB, 1] = 255 - np.floor(255*np.arange(CB)/CB)
    colorwheel[col:col+CB, 2] = 255
    col = col+CB
    # BM
    colorwheel[col:col+BM, 2] = 255
    colorwheel[col:col+BM, 0] = np.floor(255*np.arange(0,BM)/BM)
    col = col+BM
    # MR
    colorwheel[col:col+MR, 2] = 255 - np.floor(255*np.arange(MR)/MR)
    colorwheel[col:col+MR, 0] = 255
    return colorwheel


def flow_compute_color(u, v, convert_to_bgr=False):
    '''
    Applies the flow color wheel to (possibly clipped) flow components u and v.
    According to the C++ source code of Daniel Scharstein
    According to the Matlab source code of Deqing Sun
    :param u: np.ndarray, input horizontal flow
    :param v: np.ndarray, input vertical flow
    :param convert_to_bgr: bool, whether to change ordering and output BGR instead of RGB
    :return:
    
    Copied from https://github.com/tomrunia/OpticalFlow_Visualization/blob/master/flow_vis.py
    Copyright (c) 2018 Tom Runia
    '''

    flow_image = np.zeros((u.shape[0], u.shape[1], 3), np.uint8)

    colorwheel = make_colorwheel()  # shape [55x3]
    ncols = colorwheel.shape[0]

    rad = np.sqrt(np.square(u) + np.square(v))
    a = np.arctan2(-v, -u)/np.pi

    fk = (a+1) / 2*(ncols-1)
    k0 = np.floor(fk).astype(np.int32)
    k1 = k0 + 1
    k1[k1 == ncols] = 0
    f = fk - k0

    for i in range(colorwheel.shape[1]):

        tmp = colorwheel[:,i]
        col0 = tmp[k0] / 255.0
        col1 = tmp[k1] / 255.0
        col = (1-f)*col0 + f*col1

        idx = (rad <= 1)
        col[idx]  = 1 - rad[idx] * (1-col[idx])
        col[~idx] = col[~idx] * 0.75   # out of range?

        # Note the 2-i => BGR instead of RGB
        ch_idx = 2-i if convert_to_bgr else i
        flow_image[:,:,ch_idx] = np.floor(255 * col)

    return flow_image


def flow_to_color(flow_uv, clip_flow=None, convert_to_bgr=False):
    '''
    Expects a two dimensional flow image of shape [H,W,2]
    According to the C++ source code of Daniel Scharstein
    According to the Matlab source code of Deqing Sun
    :param flow_uv: np.ndarray of shape [H,W,2]
    :param clip_flow: float, maximum clipping value for flow
    :return:
    
    Copied from https://github.com/tomrunia/OpticalFlow_Visualization/blob/master/flow_vis.py
    Copyright (c) 2018 Tom Runia
    '''

    assert flow_uv.ndim == 3, 'input flow must have three dimensions'
    assert flow_uv.shape[2] == 2, 'input flow must have shape [H,W,2]'

    if clip_flow is not None:
        flow_uv = np.clip(flow_uv, 0, clip_flow)

    u = flow_uv[:,:,0]
    v = flow_uv[:,:,1]

    rad = np.sqrt(np.square(u) + np.square(v))
    rad_max = np.max(rad)

    epsilon = 1e-5
    u = u / (rad_max + epsilon)
    v = v / (rad_max + epsilon)

    return flow_compute_color(u, v, convert_to_bgr)



def show_flow( img0, img1, flow, mask=None ):
    img0 = np.asarray(img0)
    img1 = np.asarray(img1)
    if mask is None: mask = 1
    mask = np.asarray(mask)
    if mask.ndim == 2: mask = mask[:,:,None]
    assert flow.ndim == 3
    assert flow.shape[:2] == img0.shape[:2] and flow.shape[2] == 2
    
    def noticks():
      pl.xticks([])
      pl.yticks([])
    fig = pl.figure("showing correspondences")
    ax1 = pl.subplot(221)
    ax1.numaxis = 0
    pl.imshow(img0*mask)
    noticks()
    ax2 = pl.subplot(222)
    ax2.numaxis = 1
    pl.imshow(img1)
    noticks()
    
    ax = pl.subplot(212)
    ax.numaxis = 0
    flow_img = flow_to_color(np.where(np.isnan(flow), 0, flow))
    pl.imshow(flow_img * mask)
    noticks()
    
    pl.subplots_adjust(0.01, 0.01, 0.99, 0.99, wspace=0.02, hspace=0.02)
    
    def motion_notify_callback(event):
      if event.inaxes is None: return
      x,y = event.xdata, event.ydata
      ax1.lines = []
      ax2.lines = []
      try:
        x,y = int(x+0.5), int(y+0.5)
        ax1.plot(x,y,'+',ms=10,mew=2,color='blue',scalex=False,scaley=False)
        x,y = flow[y,x] + (x,y)
        ax2.plot(x,y,'+',ms=10,mew=2,color='red',scalex=False,scaley=False)
        # we redraw only the concerned axes
        renderer = fig.canvas.get_renderer()
        ax1.draw(renderer)
        ax2.draw(renderer)
        fig.canvas.blit(ax1.bbox)
        fig.canvas.blit(ax2.bbox)
      except IndexError:
        return
  
    cid_move = fig.canvas.mpl_connect('motion_notify_event',motion_notify_callback)
    print("Move your mouse over the images to show matches (ctrl-C to quit)")
    pl.show()





# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb
import torch
import torch.nn as nn
import torch.nn.functional as F


class BaseNet (nn.Module):
    """ Takes a list of images as input, and returns for each image:
        - a pixelwise descriptor
        - a pixelwise confidence
    """
    def softmax(self, ux):
        if ux.shape[1] == 1:
            x = F.softplus(ux)
            return x / (1 + x)  # for sure in [0,1], much less plateaus than softmax
        elif ux.shape[1] == 2:
            return F.softmax(ux, dim=1)[:,1:2]

    def normalize(self, x, ureliability, urepeatability):
        return dict(descriptors = F.normalize(x, p=2, dim=1),
                    repeatability = self.softmax( urepeatability ),
                    reliability = self.softmax( ureliability ))

    def forward_one(self, x):
        raise NotImplementedError()

    def forward(self, imgs, **kw):
        res = [self.forward_one(img) for img in imgs]
        # merge all dictionaries into one
        res = {k:[r[k] for r in res if k in r] for k in {k for r in res for k in r}}
        return dict(res, imgs=imgs, **kw)



class PatchNet (BaseNet):
    """ Helper class to construct a fully-convolutional network that
        extract a l2-normalized patch descriptor.
    """
    def __init__(self, inchan=3, dilated=True, dilation=1, bn=True, bn_affine=False):
        BaseNet.__init__(self)
        self.inchan = inchan
        self.curchan = inchan
        self.dilated = dilated
        self.dilation = dilation
        self.bn = bn
        self.bn_affine = bn_affine
        self.ops = nn.ModuleList([])

    def _make_bn(self, outd):
        return nn.BatchNorm2d(outd, affine=self.bn_affine)

    def _add_conv(self, outd, k=3, stride=1, dilation=1, bn=True, relu=True, k_pool = 1, pool_type='max'):
        # as in the original implementation, dilation is applied at the end of layer, so it will have impact only from next layer
        d = self.dilation * dilation
        if self.dilated: 
            conv_params = dict(padding=((k-1)*d)//2, dilation=d, stride=1)
            self.dilation *= stride
        else:
            conv_params = dict(padding=((k-1)*d)//2, dilation=d, stride=stride)
        self.ops.append( nn.Conv2d(self.curchan, outd, kernel_size=k, **conv_params) )
        if bn and self.bn: self.ops.append( self._make_bn(outd) )
        if relu: self.ops.append( nn.ReLU(inplace=True) )
        self.curchan = outd
        
        if k_pool > 1:
            if pool_type == 'avg':
                self.ops.append(torch.nn.AvgPool2d(kernel_size=k_pool))
            elif pool_type == 'max':
                self.ops.append(torch.nn.MaxPool2d(kernel_size=k_pool))
            else:
                print(f"Error, unknown pooling type {pool_type}...")
    
    def forward_one(self, x):
        assert self.ops, "You need to add convolutions first"
        for n,op in enumerate(self.ops):
            x = op(x)
        return self.normalize(x)


class L2_Net (PatchNet):
    """ Compute a 128D descriptor for all overlapping 32x32 patches.
        From the L2Net paper (CVPR'17).
    """
    def __init__(self, dim=128, **kw ):
        PatchNet.__init__(self, **kw)
        add_conv = lambda n,**kw: self._add_conv((n*dim)//128,**kw)
        add_conv(32)
        add_conv(32)
        add_conv(64, stride=2)
        add_conv(64)
        add_conv(128, stride=2)
        add_conv(128)
        add_conv(128, k=7, stride=8, bn=False, relu=False)
        self.out_dim = dim


class Quad_L2Net (PatchNet):
    """ Same than L2_Net, but replace the final 8x8 conv by 3 successive 2x2 convs.
    """
    def __init__(self, dim=128, mchan=4, relu22=False, **kw ):
        PatchNet.__init__(self, **kw)
        self._add_conv(  8*mchan)
        self._add_conv(  8*mchan)
        self._add_conv( 16*mchan, stride=2)
        self._add_conv( 16*mchan)
        self._add_conv( 32*mchan, stride=2)
        self._add_conv( 32*mchan)
        # replace last 8x8 convolution with 3 2x2 convolutions
        self._add_conv( 32*mchan, k=2, stride=2, relu=relu22)
        self._add_conv( 32*mchan, k=2, stride=2, relu=relu22)
        self._add_conv(dim, k=2, stride=2, bn=False, relu=False)
        self.out_dim = dim



class Quad_L2Net_ConfCFS (Quad_L2Net):
    """ Same than Quad_L2Net, with 2 confidence maps for repeatability and reliability.
    """
    def __init__(self, **kw ):
        Quad_L2Net.__init__(self, **kw)
        # reliability classifier
        self.clf = nn.Conv2d(self.out_dim, 2, kernel_size=1)
        # repeatability classifier: for some reasons it's a softplus, not a softmax!
        # Why? I guess it's a mistake that was left unnoticed in the code for a long time...
        self.sal = nn.Conv2d(self.out_dim, 1, kernel_size=1) 

    def forward_one(self, x):
        assert self.ops, "You need to add convolutions first"
        for op in self.ops:
            x = op(x)
        # compute the confidence maps
        ureliability = self.clf(x**2)
        urepeatability = self.sal(x**2)
        return self.normalize(x, ureliability, urepeatability)


class Fast_Quad_L2Net (PatchNet):
    """ Faster version of Quad l2 net, replacing one dilated conv with one pooling to diminish image resolution thus increase inference time
    Dilation  factors and pooling:
        1,1,1, pool2, 1,1, 2,2, 4, 8, upsample2
    """
    def __init__(self, dim=128, mchan=4, relu22=False, downsample_factor=2, **kw ):

        PatchNet.__init__(self, **kw)
        self._add_conv(  8*mchan)
        self._add_conv(  8*mchan)
        self._add_conv( 16*mchan, k_pool = downsample_factor) # added avg pooling to decrease img resolution
        self._add_conv( 16*mchan)
        self._add_conv( 32*mchan, stride=2)
        self._add_conv( 32*mchan)
        
        # replace last 8x8 convolution with 3 2x2 convolutions
        self._add_conv( 32*mchan, k=2, stride=2, relu=relu22)
        self._add_conv( 32*mchan, k=2, stride=2, relu=relu22)
        self._add_conv(dim, k=2, stride=2, bn=False, relu=False)
        
        # Go back to initial image resolution with upsampling
        self.ops.append(torch.nn.Upsample(scale_factor=downsample_factor, mode='bilinear', align_corners=False))
        
        self.out_dim = dim
        
        
class Fast_Quad_L2Net_ConfCFS (Fast_Quad_L2Net):
    """ Fast r2d2 architecture
    """
    def __init__(self, **kw ):
        Fast_Quad_L2Net.__init__(self, **kw)
        # reliability classifier
        self.clf = nn.Conv2d(self.out_dim, 2, kernel_size=1)
        
        # repeatability classifier: for some reasons it's a softplus, not a softmax!
        # Why? I guess it's a mistake that was left unnoticed in the code for a long time...
        self.sal = nn.Conv2d(self.out_dim, 1, kernel_size=1) 
        
    def forward_one(self, x):
        assert self.ops, "You need to add convolutions first"
        for op in self.ops:
            x = op(x)
        # compute the confidence maps
        ureliability = self.clf(x**2)
        urepeatability = self.sal(x**2)
        return self.normalize(x, ureliability, urepeatability)


from .basemodel import MethodType, BaseExtractor
from .detector_ensemble import EnsembleDetector, EnsembleDetectorConfig

__all__ = [
	"MethodType",
	"BaseExtractor",
	"EnsembleDetector",
	"EnsembleDetectorConfig",
]



import numpy as np
import torch
from omegaconf import OmegaConf

from easy_local_features.submodules.git_aliked.aliked import ALIKED

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..utils import ops
from .basemodel import BaseExtractor, MethodType
from typing import TypedDict


class ALIKEDConfig(TypedDict):
    model_name: str
    top_k: int
    detection_threshold: float
    force_num_keypoints: bool
    nms_radius: int


class ALIKED_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    """ALIKED baseline implementation.
        model_name: str = 'aliked-n32', Choose from ['aliked-t16', 'aliked-n16', 'aliked-n16rot', 'aliked-n32']
        top_k: int = -1, # -1 for threshold based mode, >0 for top K mode.
        scores_th: float = 0.2, # Threshold for top K = -1 mode
    """
    default_conf: ALIKEDConfig = {
        "model_name": "aliked-n16",
        "top_k": -1,
        "detection_threshold": 0.2,
        "force_num_keypoints": False,
        "nms_radius": 2,
    }

    def __init__(self, conf: ALIKEDConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)

        self.DEV = torch.device("cpu")

        self.model = ALIKED(
            {
                "model_name": conf.model_name,
                "max_num_keypoints": conf.top_k,
                "detection_threshold": conf.detection_threshold,
                "force_num_keypoints": conf.force_num_keypoints,
                "pretrained": True,
                "nms_radius": conf.nms_radius,
            }
        )

        self.model = self.model.to(self.DEV)
        self.model.eval()

        self.matcher = NearestNeighborMatcher()

    def detectAndCompute(self, img, return_dict=False):
        image = ops.prepareImage(img, batch=True).to(self.DEV)
        with torch.no_grad():
            res = self.model({"image": image})

        keypoints = res["keypoints"]
        descriptors = res["descriptors"]

        if return_dict:
            return res

        return keypoints, descriptors

    def detect(self, img):
        keypoints, descriptors = self.detectAndCompute(img)
        return keypoints

    def compute(self, img, keypoints):
        if isinstance(keypoints, np.ndarray) or isinstance(keypoints, list):
            keypoints = torch.tensor(keypoints)

        keypoints = keypoints.to(self.DEV)

        if len(keypoints) == 0:
            return torch.zeros(0, 128).to(self.DEV)

        if len(keypoints.shape) == 2:
            keypoints = keypoints.unsqueeze(0)

        image = ops.prepareImage(img, batch=True).to(self.DEV)
        with torch.no_grad():
            res = self.model.forward_desc({"image": image}, keypoints)

        return res["descriptors"]

    def to(self, device):
        self.model.to(device)
        self.DEV = device

    @property
    def has_detector(self):
        return True


if __name__ == "__main__":
    from easy_local_features.utils import io, ops

    method = ALIKED_baseline(
        {
            "model_name": "aliked-n16",
            "top_k": 128,
            "detection_threshold": 0.2,
            "force_num_keypoints": False,
            "nms_radius": 2,
        }
    )

    img0 = io.fromPath("test/assets/megadepth0.jpg")

    kpts = method.detect(img0)
    desc = method.compute(img0, kpts)

    kpts2, desc2 = method.detectAndCompute(img0)

    assert torch.allclose(kpts, kpts2)

    # import pdb; pdb.set_trace()
    print(desc)
    print(desc2)
    assert torch.allclose(desc, desc2, atol=1e-5)



import torch
from omegaconf import OmegaConf

from kornia.feature import LightGlue
from ..feature.basemodel import BaseExtractor
from .. import getExtractor
from ..utils import io, vis, ops

class LightGlue_baseline(BaseExtractor):
    defalut_conf = {
        "features": "superpoint",
        "top_k": 2048,
    }
    
    variations = {
        "features": list(LightGlue.features.keys())
    }
    
    def __init__(self, conf={}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.defalut_conf), conf)
        features = conf.get("features", self.defalut_conf["features"])
        
        assert features in LightGlue.features, f"Invalid feature {features}. Available features: {list(LightGlue.features.keys())}"
        
        self.device = torch.device("cpu")
        self.matcher = LightGlue(features).to(self.device)

        self._extractor = None

    @property
    def extractor(self):
        if self._extractor is None:
            if self.conf["features"] == "dedodeg":
                tmp_conf = self.conf.copy()
                tmp_conf["descriptor_weights"] = "G-upright"
                self._extractor = getExtractor("dedode", tmp_conf)
            elif self.conf["features"] == "dedodeb":
                tmp_conf = self.conf.copy()
                tmp_conf["descriptor_weights"] = "B-upright"
                self._extractor = getExtractor("dedode", tmp_conf)
            else:
                self._extractor = getExtractor(self.defalut_conf["features"], self.conf)
            self._extractor.to(self.device)
        return self._extractor        

    def detectAndCompute(self, image, return_dict=False):
        return self.extractor.detectAndCompute(image, return_dict=return_dict)

    def detect(self, image):
        return self.extractor.detect(image)
    
    def compute(self, image, keypoints):
        return self.extractor.compute(image, keypoints)
    
    @property
    def has_detector(self):
        return self.extractor.has_detector

    def match(self, image0, image1):        
        data0 = self.detectAndCompute(image0, return_dict=True)
        data1 = self.detectAndCompute(image1, return_dict=True)
        
        data = {
            "keypoints0": data0["keypoints"],
            "keypoints1": data1["keypoints"],
            "descriptors0": data0["descriptors"],
            "descriptors1": data1["descriptors"],
            "image0": image0,
            "image1": image1,
        }

        return self.match_cached(data)

    def match_cached(self, data):
        """
        data: dict with keys:
        data = 
        {
            "keypoints0": keypoints0,
            "keypoints1": keypoints1,
            "descriptors0": descriptors0,
            "descriptors1": descriptors1,
            "image0": image0,
            "image1": image1,
        }
        """

        keypoints0 = data["keypoints0"]
        keypoints1 = data["keypoints1"]
        descriptors0 = data["descriptors0"]
        descriptors1 = data["descriptors1"]

        input = {
            "image0": {
                "keypoints": keypoints0,
                "descriptors": descriptors0,
                "image": data["image0"] if isinstance(data["image0"], torch.Tensor) else ops.prepareImage(data["image0"]),
            },
            "image1": {
                "keypoints": keypoints1,
                "descriptors": descriptors1,
                "image": data["image1"] if isinstance(data["image1"], torch.Tensor) else ops.prepareImage(data["image1"]),
            },
        }
        
        with torch.no_grad():
            out = self.matcher(input)

        matches = out["matches"][0].to(self.device)
        mscores = out["matching_scores0"][0].to(self.device)

        mkpts0 = keypoints0[0, matches[:, 0]]
        mkpts1 = keypoints1[0, matches[:, 1]]
        # Ensure keypoints and matches are detached and on CPU for output
        return {
            "keypoints0": keypoints0.detach().cpu(),
            "keypoints1": keypoints1.detach().cpu(),
            "descriptors0": descriptors0.detach().cpu() if isinstance(descriptors0, torch.Tensor) else descriptors0,
            "descriptors1": descriptors1.detach().cpu() if isinstance(descriptors1, torch.Tensor) else descriptors1,
            "matches": matches.detach().cpu(),
            "scores": mscores.detach().cpu(),
            "mkpts0": mkpts0.detach().cpu(),
            "mkpts1": mkpts1.detach().cpu(),
        }
    
    def to(self, device):
        self.matcher = self.matcher.to(device)
        if self._extractor is not None:
            self._extractor.to(device)
        self.device = device
        return self


from typing import Optional, TypedDict

import torch
from torch import nn
from omegaconf import OmegaConf

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..utils import ops
from .basemodel import BaseExtractor, MethodType

"""
# %BANNER_BEGIN%
# ---------------------------------------------------------------------
# %COPYRIGHT_BEGIN%
#
#  Magic Leap, Inc. ("COMPANY") CONFIDENTIAL
#
#  Unpublished Copyright (c) 2020
#  Magic Leap, Inc., All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property
# of COMPANY. The intellectual and technical concepts contained herein
# are proprietary to COMPANY and may be covered by U.S. and Foreign
# Patents, patents in process, and are protected by trade secret or
# copyright law.  Dissemination of this information or reproduction of
# this material is strictly forbidden unless prior written permission is
# obtained from COMPANY.  Access to the source code contained herein is
# hereby forbidden to anyone except current COMPANY employees, managers
# or contractors who have executed Confidentiality and Non-disclosure
# agreements explicitly covering such access.
#
# The copyright notice above does not evidence any actual or intended
# publication or disclosure  of  this source code, which includes
# information that is confidential and/or proprietary, and is a trade
# secret, of  COMPANY.   ANY REPRODUCTION, MODIFICATION, DISTRIBUTION,
# PUBLIC  PERFORMANCE, OR PUBLIC DISPLAY OF OR THROUGH USE  OF THIS
# SOURCE CODE  WITHOUT THE EXPRESS WRITTEN CONSENT OF COMPANY IS
# STRICTLY PROHIBITED, AND IN VIOLATION OF APPLICABLE LAWS AND
# INTERNATIONAL TREATIES.  THE RECEIPT OR POSSESSION OF  THIS SOURCE
# CODE AND/OR RELATED INFORMATION DOES NOT CONVEY OR IMPLY ANY RIGHTS
# TO REPRODUCE, DISCLOSE OR DISTRIBUTE ITS CONTENTS, OR TO MANUFACTURE,
# USE, OR SELL ANYTHING THAT IT  MAY DESCRIBE, IN WHOLE OR IN PART.
#
# %COPYRIGHT_END%
# ----------------------------------------------------------------------
# %AUTHORS_BEGIN%
#
#  Originating Authors: Paul-Edouard Sarlin
#
# %AUTHORS_END%
# --------------------------------------------------------------------*/
# %BANNER_END%

Described in:
    SuperPoint: Self-Supervised Interest Point Detection and Description,
    Daniel DeTone, Tomasz Malisiewicz, Andrew Rabinovich, CVPRW 2018.

Original code: github.com/MagicLeapResearch/SuperPointPretrainedNetwork

Adapted by Philipp Lindenberger (Phil26AT)
"""


def pad_to_length(
    x,
    length: int,
    pad_dim: int = -2,
    mode: str = "zeros",  # zeros, ones, random, random_c
    bounds: tuple[int] = (None, None),
):
    shape = list(x.shape)
    d = x.shape[pad_dim]
    assert d <= length
    if d == length:
        return x
    shape[pad_dim] = length - d

    low, high = bounds

    if mode == "zeros":
        xn = torch.zeros(*shape, device=x.device, dtype=x.dtype)
    elif mode == "ones":
        xn = torch.ones(*shape, device=x.device, dtype=x.dtype)
    elif mode == "random":
        low = low if low is not None else x.min()
        high = high if high is not None else x.max()
        xn = torch.empty(*shape, device=x.device).uniform_(low, high)
    elif mode == "random_c":
        low, high = bounds  # we use the bounds as fallback for empty seq.
        xn = torch.cat(
            [
                torch.empty(*shape[:-1], 1, device=x.device).uniform_(
                    x[..., i].min() if d > 0 else low,
                    x[..., i].max() if d > 0 else high,
                )
                for i in range(shape[-1])
            ],
            dim=-1,
        )
    else:
        raise ValueError(mode)
    return torch.cat([x, xn], dim=pad_dim)


def pad_and_stack(
    sequences: list[torch.Tensor],
    length: Optional[int] = None,
    pad_dim: int = -2,
    **kwargs,
):
    if length is None:
        length = max([x.shape[pad_dim] for x in sequences])

    y = torch.stack([pad_to_length(x, length, pad_dim, **kwargs) for x in sequences], 0)
    return y


def simple_nms(scores, radius):
    """Perform non maximum suppression on the heatmap using max-pooling.
    This method does not suppress contiguous points that have the same score.
    Args:
        scores: the score heatmap of size `(B, H, W)`.
        radius: an integer scalar, the radius of the NMS window.
    """

    def max_pool(x):
        return torch.nn.functional.max_pool2d(x, kernel_size=radius * 2 + 1, stride=1, padding=radius)

    zeros = torch.zeros_like(scores)
    max_mask = scores == max_pool(scores)
    for _ in range(2):
        supp_mask = max_pool(max_mask.float()) > 0
        supp_scores = torch.where(supp_mask, zeros, scores)
        new_max_mask = supp_scores == max_pool(supp_scores)
        max_mask = max_mask | (new_max_mask & (~supp_mask))
    return torch.where(max_mask, scores, zeros)


def top_k_keypoints(keypoints, scores, k):
    if k >= len(keypoints):
        return keypoints, scores
    scores, indices = torch.topk(scores, k, dim=0, sorted=True)
    return keypoints[indices], scores


def sample_k_keypoints(keypoints, scores, k):
    if k >= len(keypoints):
        return keypoints, scores
    indices = torch.multinomial(scores, k, replacement=False)
    return keypoints[indices], scores[indices]


def soft_argmax_refinement(keypoints, scores, radius: int):
    width = 2 * radius + 1
    sum_ = torch.nn.functional.avg_pool2d(scores[:, None], width, 1, radius, divisor_override=1)
    ar = torch.arange(-radius, radius + 1).to(scores)
    kernel_x = ar[None].expand(width, -1)[None, None]
    dx = torch.nn.functional.conv2d(scores[:, None], kernel_x, padding=radius)
    dy = torch.nn.functional.conv2d(scores[:, None], kernel_x.transpose(2, 3), padding=radius)
    dydx = torch.stack([dy[:, 0], dx[:, 0]], -1) / sum_[:, 0, :, :, None]
    refined_keypoints = []
    for i, kpts in enumerate(keypoints):
        delta = dydx[i][tuple(kpts.t())]
        refined_keypoints.append(kpts.float() + delta)
    return refined_keypoints


# Legacy (broken) sampling of the descriptors
def sample_descriptors(keypoints, descriptors, s):
    b, c, h, w = descriptors.shape
    keypoints = keypoints - s / 2 + 0.5
    keypoints /= torch.tensor(
        [(w * s - s / 2 - 0.5), (h * s - s / 2 - 0.5)],
    ).to(keypoints)[None]
    keypoints = keypoints * 2 - 1  # normalize to (-1, 1)
    args = {"align_corners": True} if torch.__version__ >= "1.3" else {}
    descriptors = torch.nn.functional.grid_sample(descriptors, keypoints.view(b, 1, -1, 2), mode="bilinear", **args)
    descriptors = torch.nn.functional.normalize(descriptors.reshape(b, c, -1), p=2, dim=1)
    return descriptors


# The original keypoint sampling is incorrect. We patch it here but
# keep the original one above for legacy.
def sample_descriptors_fix_sampling(keypoints, descriptors, s: int = 8):
    """Interpolate descriptors at keypoint locations"""
    b, c, h, w = descriptors.shape
    keypoints = keypoints / (keypoints.new_tensor([w, h]) * s)
    keypoints = keypoints * 2 - 1  # normalize to (-1, 1)
    descriptors = torch.nn.functional.grid_sample(
        descriptors, keypoints.view(b, 1, -1, 2), mode="bilinear", align_corners=False
    )
    descriptors = torch.nn.functional.normalize(descriptors.reshape(b, c, -1), p=2, dim=1)
    return descriptors


class Net(nn.Module):
    def __init__(self, conf):
        super().__init__()
        self.conf = conf
        self.relu = nn.ReLU(inplace=True)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        c1, c2, c3, c4, c5 = 64, 64, 128, 128, 256
        descriptor_dim = 256

        self.conv1a = nn.Conv2d(1, c1, kernel_size=3, stride=1, padding=1)
        self.conv1b = nn.Conv2d(c1, c1, kernel_size=3, stride=1, padding=1)
        self.conv2a = nn.Conv2d(c1, c2, kernel_size=3, stride=1, padding=1)
        self.conv2b = nn.Conv2d(c2, c2, kernel_size=3, stride=1, padding=1)
        self.conv3a = nn.Conv2d(c2, c3, kernel_size=3, stride=1, padding=1)
        self.conv3b = nn.Conv2d(c3, c3, kernel_size=3, stride=1, padding=1)
        self.conv4a = nn.Conv2d(c3, c4, kernel_size=3, stride=1, padding=1)
        self.conv4b = nn.Conv2d(c4, c4, kernel_size=3, stride=1, padding=1)

        self.convPa = nn.Conv2d(c4, c5, kernel_size=3, stride=1, padding=1)
        self.convPb = nn.Conv2d(c5, 65, kernel_size=1, stride=1, padding=0)

        self.convDa = nn.Conv2d(c4, c5, kernel_size=3, stride=1, padding=1)
        self.convDb = nn.Conv2d(c5, descriptor_dim, kernel_size=1, stride=1, padding=0)

    def forward(self, data, dense_outputs=False):
        image = data["image"]
        if image.shape[1] == 3:  # RGB
            scale = image.new_tensor([0.299, 0.587, 0.114]).view(1, 3, 1, 1)
            image = (image * scale).sum(1, keepdim=True)

        # Shared Encoder
        x = self.relu(self.conv1a(image))
        x = self.relu(self.conv1b(x))
        x = self.pool(x)
        x = self.relu(self.conv2a(x))
        x = self.relu(self.conv2b(x))
        x = self.pool(x)
        x = self.relu(self.conv3a(x))
        x = self.relu(self.conv3b(x))
        x = self.pool(x)
        x = self.relu(self.conv4a(x))
        x = self.relu(self.conv4b(x))

        pred = {}

        # Compute the dense keypoint scores
        cPa = self.relu(self.convPa(x))
        scores = self.convPb(cPa)
        scores = torch.nn.functional.softmax(scores, 1)[:, :-1]
        b, c, h, w = scores.shape
        scores = scores.permute(0, 2, 3, 1).reshape(b, h, w, 8, 8)
        scores = scores.permute(0, 1, 3, 2, 4).reshape(b, h * 8, w * 8)
        pred["keypoint_scores"] = dense_scores = scores

        # Compute the dense descriptors
        cDa = self.relu(self.convDa(x))
        dense_desc = self.convDb(cDa)
        dense_desc = torch.nn.functional.normalize(dense_desc, p=2, dim=1)
        pred["descriptors"] = dense_desc

        if self.conf.sparse_outputs:
            scores = simple_nms(scores, self.conf.nms_radius)

            # Discard keypoints near the image borders
            if self.conf.remove_borders:
                scores[:, : self.conf.remove_borders] = -1
                scores[:, :, : self.conf.remove_borders] = -1
                if "image_size" in data:
                    for i in range(scores.shape[0]):
                        w, h = data["image_size"][i]
                        scores[i, int(h.item()) - self.conf.remove_borders :] = -1
                        scores[i, :, int(w.item()) - self.conf.remove_borders :] = -1
                else:
                    scores[:, -self.conf.remove_borders :] = -1
                    scores[:, :, -self.conf.remove_borders :] = -1

            # Extract keypoints
            best_kp = torch.where(scores > self.conf.detection_threshold)
            scores = scores[best_kp]

            # Separate into batches
            keypoints = [torch.stack(best_kp[1:3], dim=-1)[best_kp[0] == i] for i in range(b)]
            scores = [scores[best_kp[0] == i] for i in range(b)]

            # Keep the k keypoints with highest score
            max_kps = self.conf.top_k

            # Keep the k keypoints with highest score
            if max_kps > 0:
                keypoints, scores = list(zip(*[top_k_keypoints(k, s, max_kps) for k, s in zip(keypoints, scores)]))
                keypoints, scores = list(keypoints), list(scores)

            if self.conf["refinement_radius"] > 0:
                keypoints = soft_argmax_refinement(keypoints, dense_scores, self.conf["refinement_radius"])

            # Convert (h, w) to (x, y)
            keypoints = [torch.flip(k, [1]).float() for k in keypoints]

            keypoints = torch.stack(keypoints, 0)
            scores = torch.stack(scores, 0)

            # Extract descriptors
            if (len(keypoints) == 1) or self.conf.force_num_keypoints:
                # Batch sampling of the descriptors
                if self.conf.legacy_sampling:
                    desc = sample_descriptors(keypoints, dense_desc, 8)
                else:
                    desc = sample_descriptors_fix_sampling(keypoints, dense_desc, 8)
            else:
                if self.conf.legacy_sampling:
                    desc = [sample_descriptors(k[None], d[None], 8)[0] for k, d in zip(keypoints, dense_desc)]
                else:
                    desc = [
                        sample_descriptors_fix_sampling(k[None], d[None], 8)[0] for k, d in zip(keypoints, dense_desc)
                    ]

            pred = {
                "keypoints": keypoints + 0.5,
                "keypoint_scores": scores,
                "descriptors": desc.transpose(-1, -2),
            }

            if self.conf.dense_outputs or dense_outputs:
                pred["dense_descriptors"] = dense_desc

        return pred


class SuperPointConfig(TypedDict):
    top_k: int
    sparse_outputs: bool
    dense_outputs: bool
    nms_radius: int
    refinement_radius: int
    detection_threshold: float
    remove_borders: int
    legacy_sampling: bool

class SuperPoint_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    default_conf: SuperPointConfig = {
        "top_k": -1,
        "sparse_outputs": True,
        "dense_outputs": False,
        "nms_radius": 4,
        "refinement_radius": 0,
        "detection_threshold": 0.005,
        "remove_borders": 4,
        "legacy_sampling": True,  # True to use the old broken sampling
    }

    checkpoint_url = (
        "https://github.com/magicleap/SuperGluePretrainedNetwork/raw/master/models/weights/superpoint_v1.pth"  # noqa: E501
    )

    def __init__(self, conf: SuperPointConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        self.device = torch.device("cpu")
        self.matcher = NearestNeighborMatcher()
        self.model = Net(conf)
        self.model.load_state_dict(torch.hub.load_state_dict_from_url(str(self.checkpoint_url)), strict=False)
        self.model.eval()

    def detectAndCompute(self, img, return_dict=False, sparse_outputs=False):
        img = ops.prepareImage(img).to(self.device)
        pred = self.model({"image": img})

        if return_dict:
            return pred

        return pred["keypoints"], pred["descriptors"]

    def detect(self, img):
        return self.detectAndCompute(img)[0]

    def compute(self, img, keypoints):
        img = ops.prepareImage(img).to(self.device)
        pred = self.model({"image": img}, dense_outputs=True)
        dense_desc = pred["dense_descriptors"]
        keypoints = keypoints - 0.5

        if self.conf.legacy_sampling:
            desc = sample_descriptors(keypoints, dense_desc, 8)
        else:
            desc = sample_descriptors_fix_sampling(keypoints, dense_desc, 8)

        return desc.transpose(-1, -2)

    def to(self, device):
        self.model = self.model.to(device)
        self.device = device

    @property
    def has_detector(self):
        return True


if __name__ == "__main__":
    from easy_local_features.utils import io, ops

    method = SuperPoint_baseline({"legacy_sampling": False})

    img0 = io.fromPath("test/assets/megadepth0.jpg")

    kpts = method.detect(img0)
    desc = method.compute(img0, kpts)

    kpts2, desc2 = method.detectAndCompute(img0)

    assert torch.allclose(kpts, kpts2)

    # import pdb; pdb.set_trace()
    print(desc)
    print(desc2)
    assert torch.allclose(desc, desc2, atol=1e-5)



# %BANNER_BEGIN%
# ---------------------------------------------------------------------
# %COPYRIGHT_BEGIN%
#
#  Magic Leap, Inc. ("COMPANY") CONFIDENTIAL
#
#  Unpublished Copyright (c) 2020
#  Magic Leap, Inc., All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property
# of COMPANY. The intellectual and technical concepts contained herein
# are proprietary to COMPANY and may be covered by U.S. and Foreign
# Patents, patents in process, and are protected by trade secret or
# copyright law.  Dissemination of this information or reproduction of
# this material is strictly forbidden unless prior written permission is
# obtained from COMPANY.  Access to the source code contained herein is
# hereby forbidden to anyone except current COMPANY employees, managers
# or contractors who have executed Confidentiality and Non-disclosure
# agreements explicitly covering such access.
#
# The copyright notice above does not evidence any actual or intended
# publication or disclosure  of  this source code, which includes
# information that is confidential and/or proprietary, and is a trade
# secret, of  COMPANY.   ANY REPRODUCTION, MODIFICATION, DISTRIBUTION,
# PUBLIC  PERFORMANCE, OR PUBLIC DISPLAY OF OR THROUGH USE  OF THIS
# SOURCE CODE  WITHOUT THE EXPRESS WRITTEN CONSENT OF COMPANY IS
# STRICTLY PROHIBITED, AND IN VIOLATION OF APPLICABLE LAWS AND
# INTERNATIONAL TREATIES.  THE RECEIPT OR POSSESSION OF  THIS SOURCE
# CODE AND/OR RELATED INFORMATION DOES NOT CONVEY OR IMPLY ANY RIGHTS
# TO REPRODUCE, DISCLOSE OR DISTRIBUTE ITS CONTENTS, OR TO MANUFACTURE,
# USE, OR SELL ANYTHING THAT IT  MAY DESCRIBE, IN WHOLE OR IN PART.
#
# %COPYRIGHT_END%
# ----------------------------------------------------------------------
# %AUTHORS_BEGIN%
#
#  Originating Authors: Paul-Edouard Sarlin
#
# %AUTHORS_END%
# --------------------------------------------------------------------*/
# %BANNER_END%

from pathlib import Path
import torch
from torch import nn

def simple_nms(scores, nms_radius: int):
    """ Fast Non-maximum suppression to remove nearby points """
    assert(nms_radius >= 0)

    def max_pool(x):
        return torch.nn.functional.max_pool2d(
            x, kernel_size=nms_radius*2+1, stride=1, padding=nms_radius)

    zeros = torch.zeros_like(scores)
    max_mask = scores == max_pool(scores)
    for _ in range(2):
        supp_mask = max_pool(max_mask.float()) > 0
        supp_scores = torch.where(supp_mask, zeros, scores)
        new_max_mask = supp_scores == max_pool(supp_scores)
        max_mask = max_mask | (new_max_mask & (~supp_mask))
    return torch.where(max_mask, scores, zeros)


def remove_borders(keypoints, scores, border: int, height: int, width: int):
    """ Removes keypoints too close to the border """
    mask_h = (keypoints[:, 0] >= border) & (keypoints[:, 0] < (height - border))
    mask_w = (keypoints[:, 1] >= border) & (keypoints[:, 1] < (width - border))
    mask = mask_h & mask_w
    return keypoints[mask], scores[mask]


def top_k_keypoints(keypoints, scores, k: int):
    if k >= len(keypoints):
        return keypoints, scores
    scores, indices = torch.topk(scores, k, dim=0)
    return keypoints[indices], scores


def sample_descriptors(keypoints, descriptors, s: int = 8):
    """ Interpolate descriptors at keypoint locations """
    b, c, h, w = descriptors.shape
    keypoints = keypoints - s / 2 + 0.5
    keypoints /= torch.tensor([(w*s - s/2 - 0.5), (h*s - s/2 - 0.5)],
                              ).to(keypoints)[None]
    keypoints = keypoints*2 - 1  # normalize to (-1, 1)
    args = {'align_corners': True} if torch.__version__ >= '1.3' else {}
    descriptors = torch.nn.functional.grid_sample(
        descriptors, keypoints.view(b, 1, -1, 2), mode='bilinear', **args)
    descriptors = torch.nn.functional.normalize(
        descriptors.reshape(b, c, -1), p=2, dim=1)
    return descriptors


class SuperPoint(nn.Module):
    """SuperPoint Convolutional Detector and Descriptor

    SuperPoint: Self-Supervised Interest Point Detection and
    Description. Daniel DeTone, Tomasz Malisiewicz, and Andrew
    Rabinovich. In CVPRW, 2019. https://arxiv.org/abs/1712.07629

    """
    default_config = {
        'descriptor_dim': 256,
        'nms_radius': 4,
        'keypoint_threshold': 0.005,
        'max_keypoints': -1,
        'remove_borders': 4,
        'model_path': None,
    }

    def __init__(self, config):
        super().__init__()
        self.config = {**self.default_config, **config}

        self.relu = nn.ReLU(inplace=True)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        c1, c2, c3, c4, c5 = 64, 64, 128, 128, 256

        self.conv1a = nn.Conv2d(1, c1, kernel_size=3, stride=1, padding=1)
        self.conv1b = nn.Conv2d(c1, c1, kernel_size=3, stride=1, padding=1)
        self.conv2a = nn.Conv2d(c1, c2, kernel_size=3, stride=1, padding=1)
        self.conv2b = nn.Conv2d(c2, c2, kernel_size=3, stride=1, padding=1)
        self.conv3a = nn.Conv2d(c2, c3, kernel_size=3, stride=1, padding=1)
        self.conv3b = nn.Conv2d(c3, c3, kernel_size=3, stride=1, padding=1)
        self.conv4a = nn.Conv2d(c3, c4, kernel_size=3, stride=1, padding=1)
        self.conv4b = nn.Conv2d(c4, c4, kernel_size=3, stride=1, padding=1)

        self.convPa = nn.Conv2d(c4, c5, kernel_size=3, stride=1, padding=1)
        self.convPb = nn.Conv2d(c5, 65, kernel_size=1, stride=1, padding=0)

        self.convDa = nn.Conv2d(c4, c5, kernel_size=3, stride=1, padding=1)
        self.convDb = nn.Conv2d(
            c5, self.config['descriptor_dim'],
            kernel_size=1, stride=1, padding=0)

        path = config['model_path']
        self.load_state_dict(torch.load(path))

        mk = self.config['max_keypoints']
        if mk == 0 or mk < -1:
            raise ValueError('\"max_keypoints\" must be positive or \"-1\"')

    def forward(self, data):
        """ Compute keypoints, scores, descriptors for image """
        # Shared Encoder
        x = self.relu(self.conv1a(data['image']))
        x = self.relu(self.conv1b(x))
        x = self.pool(x)
        x = self.relu(self.conv2a(x))
        x = self.relu(self.conv2b(x))
        x = self.pool(x)
        x = self.relu(self.conv3a(x))
        x = self.relu(self.conv3b(x))
        x = self.pool(x)
        x = self.relu(self.conv4a(x))
        x = self.relu(self.conv4b(x))

        # Compute the dense keypoint scores
        cPa = self.relu(self.convPa(x))
        scores = self.convPb(cPa)
        scores = torch.nn.functional.softmax(scores, 1)[:, :-1]
        b, _, h, w = scores.shape
        scores = scores.permute(0, 2, 3, 1).reshape(b, h, w, 8, 8)
        scores = scores.permute(0, 1, 3, 2, 4).reshape(b, h*8, w*8)
        scores = simple_nms(scores, self.config['nms_radius'])

        # Extract keypoints
        keypoints = [
            torch.nonzero(s > self.config['keypoint_threshold'])
            for s in scores]
        scores = [s[tuple(k.t())] for s, k in zip(scores, keypoints)]

        # Discard keypoints near the image borders
        keypoints, scores = list(zip(*[
            remove_borders(k, s, self.config['remove_borders'], h*8, w*8)
            for k, s in zip(keypoints, scores)]))

        # Keep the k keypoints with highest score
        if self.config['max_keypoints'] >= 0:
            keypoints, scores = list(zip(*[
                top_k_keypoints(k, s, self.config['max_keypoints'])
                for k, s in zip(keypoints, scores)]))

        # Convert (h, w) to (x, y)
        keypoints = [torch.flip(k, [1]).float() for k in keypoints]

        # Compute the dense descriptors
        cDa = self.relu(self.convDa(x))
        descriptors = self.convDb(cDa)
        descriptors = torch.nn.functional.normalize(descriptors, p=2, dim=1)

        # Extract descriptors
        descriptors = [sample_descriptors(k[None], d[None], 8)[0]
                       for k, d in zip(keypoints, descriptors)]

        return {
            'keypoints': keypoints,
            'scores': scores,
            'descriptors': descriptors,
        }



import os

import torch
from omegaconf import OmegaConf

from easy_local_features.submodules.git_relf.descriptor_utils import DescGroupPoolandNorm
from easy_local_features.submodules.git_relf.model import RELF_model

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..utils import ops
from ..utils.download import downloadModel
from .basemodel import BaseExtractor, MethodType

try:
    import mmcv
except ImportError:
    raise ImportError(
        "Please install `mmcv` to use RELF. I know, it's a pain, but I'm not putting it in the requirements."
    )

available_models = [
    "re_resnet",
    "E2SFCNN",
    "E2SFCNN_QUOT",
    "EXP",
    "CNN",
    "wrn16_8_stl",
    "e2wrn16_8_stl",
    "e2wrn28_10",
    "e2wrn28_7",
    "e2wrn28_10R",
    "e2wrn28_7R",
    "e2wrn10_8R",
    "e2wrn16_8R_stl",
]

# class ReFTDescriptor:
#     def __init__(self, model:RELF_model):

#         self.model = model
#         self.pool_and_norm = DescGroupPoolandNorm(model.args)

#     def __call__(self, image, kpts):
#         desc = self.model(image, kpts)

#         ## kpts torch.tensor ([B, K, 2]), desc torch.tensor ([B, K, CG])
#         k1, d1 = self.pool_and_norm.desc_pool_and_norm_infer(kpts, desc)

#         return k1, d1


class RELF_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DESCRIPTOR_ONLY
    """'
    Model name:
     're_resnet': Rotation-Equivariant ResNet-18
     'e2wrn16_8R_stl': Rotation-Equivariant Wide-ResNet-16-8
    """

    default_conf = {
        "model": "re_resnet",
        "num_group": 16,
        "channels": 64,
        "candidate": "top1",  # 'topX' or [0,1] float,
        "orientation": 16,
        "top_k": 4096,
    }

    def __init__(self, conf={}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        assert conf.model in available_models, f"Model {conf.model} not available. Choose from {available_models}"

        os.environ["Orientation"] = str(conf.orientation)
        self.max_kps = conf.top_k
        self.device = torch.device("cpu")
        self.net = RELF_model(self.conf)

        if conf.model == "re_resnet":
            url = "https://github.com/felipecadar/easy-local-features-baselines/releases/download/RELF-weights/best_model.pt"
            cache_path = downloadModel("RELF", "re_resnet", url)
        elif conf.model == "e2wrn16_8R_stl":
            url = "https://github.com/felipecadar/easy-local-features-baselines/releases/download/RELF-weights/best_model_wrn.pt"
            cache_path = downloadModel("RELF", "e2wrn16_8R_stl", url)
        else:
            raise NotImplementedError(f"Model {conf.model}: weights not available.")

        weights = torch.load(cache_path, map_location=self.device)
        self.net.train()
        self.net.load_state_dict(weights)

        self.pool_and_norm = DescGroupPoolandNorm(conf)
        self.matcher = NearestNeighborMatcher(
            {
                "ratio_thresh": None,
                "distance_thresh": None,
                "mutual_check": True,
                "loss": None,
                "normalize": False,
            }
        )

        self.net.eval()
        self.net.to(self.device)

    def detectAndCompute(self, image, return_dict=False):
        raise NotImplementedError

    def detect(self, image):
        raise NotImplementedError

    def compute(self, img, kps):
        image = ops.prepareImage(img, imagenet=True).to(self.device)

        with torch.no_grad():
            desc = self.net(image, kps)
            kps, desc = self.pool_and_norm.desc_pool_and_norm_infer(kps, desc)

        return kps, desc

    def to(self, device):
        self.net.to(device)
        self.device = device
        return self

    @property
    def has_detector(self):
        return False


if __name__ == "__main__":
    from easy_local_features.feature.baseline_superpoint import SuperPoint_baseline
    from easy_local_features.utils import io

    relf = RELF_baseline()
    sp = SuperPoint_baseline(
        {
            "keypoint_threshold": 0,
            "max_keypoints": 1024,
            "force_num_keypoints": True,
        }
    )

    batch_size = 4
    imgs = io.fromPath("test/assets/megadepth0.jpg").repeat(batch_size, 1, 1, 1)

    kps = sp.detect(imgs)
    kps, desc = relf.compute(imgs, kps)

    print(kps.shape, desc.shape)



import torch
from torch import nn
import torch.nn.functional as F


# coordinates system
#  ------------------------------>  [ x: range=-1.0~1.0; w: range=0~W ]
#  | -----------------------------
#  | |                           |
#  | |                           |
#  | |                           |
#  | |         image             |
#  | |                           |
#  | |                           |
#  | |                           |
#  | |---------------------------|
#  v
# [ y: range=-1.0~1.0; h: range=0~H ]

def simple_nms(scores, nms_radius: int):
    """ Fast Non-maximum suppression to remove nearby points """
    assert (nms_radius >= 0)

    def max_pool(x):
        return torch.nn.functional.max_pool2d(
            x, kernel_size=nms_radius * 2 + 1, stride=1, padding=nms_radius)

    zeros = torch.zeros_like(scores)
    max_mask = scores == max_pool(scores)

    for _ in range(2):
        supp_mask = max_pool(max_mask.float()) > 0
        supp_scores = torch.where(supp_mask, zeros, scores)
        new_max_mask = supp_scores == max_pool(supp_scores)
        max_mask = max_mask | (new_max_mask & (~supp_mask))
    return torch.where(max_mask, scores, zeros)


def sample_descriptor(descriptor_map, kpts, bilinear_interp=False):
    """
    :param descriptor_map: BxCxHxW
    :param kpts: list, len=B, each is Nx2 (keypoints) [h,w]
    :param bilinear_interp: bool, whether to use bilinear interpolation
    :return: descriptors: list, len=B, each is NxD
    """
    batch_size, channel, height, width = descriptor_map.shape

    descriptors = []
    for index in range(batch_size):
        kptsi = kpts[index]  # Nx2,(x,y)

        if bilinear_interp:
            descriptors_ = torch.nn.functional.grid_sample(descriptor_map[index].unsqueeze(0), kptsi.view(1, 1, -1, 2),
                                                           mode='bilinear', align_corners=True)[0, :, 0, :]  # CxN
        else:
            kptsi = (kptsi + 1) / 2 * kptsi.new_tensor([[width - 1, height - 1]])
            kptsi = kptsi.long()
            descriptors_ = descriptor_map[index, :, kptsi[:, 1], kptsi[:, 0]]  # CxN

        descriptors_ = torch.nn.functional.normalize(descriptors_, p=2, dim=0)
        descriptors.append(descriptors_.t())

    return descriptors


class DKD(nn.Module):
    def __init__(self, radius=2, top_k=0, scores_th=0.2, n_limit=20000):
        """
        Args:
            radius: soft detection radius, kernel size is (2 * radius + 1)
            top_k: top_k > 0: return top k keypoints
            scores_th: top_k <= 0 threshold mode:  scores_th > 0: return keypoints with scores>scores_th
                                                   else: return keypoints with scores > scores.mean()
            n_limit: max number of keypoint in threshold mode
        """
        super().__init__()
        self.radius = radius
        self.top_k = top_k
        self.scores_th = scores_th
        self.n_limit = n_limit
        self.kernel_size = 2 * self.radius + 1
        self.temperature = 0.1  # tuned temperature
        self.unfold = nn.Unfold(kernel_size=self.kernel_size, padding=self.radius)

        # local xy grid
        x = torch.linspace(-self.radius, self.radius, self.kernel_size)
        # (kernel_size*kernel_size) x 2 : (w,h)
        self.hw_grid = torch.stack(torch.meshgrid([x, x], indexing="ij")).view(2, -1).t()[:, [1, 0]]

    def detect_keypoints(self, scores_map, sub_pixel=True):
        b, c, h, w = scores_map.shape
        scores_nograd = scores_map.detach()
        # nms_scores = simple_nms(scores_nograd, self.radius)
        nms_scores = simple_nms(scores_nograd, 2)

        # remove border
        nms_scores[:, :, :self.radius + 1, :] = 0
        nms_scores[:, :, :, :self.radius + 1] = 0
        nms_scores[:, :, h - self.radius:, :] = 0
        nms_scores[:, :, :, w - self.radius:] = 0

        # detect keypoints without grad
        if self.top_k > 0:
            topk = torch.topk(nms_scores.view(b, -1), self.top_k)
            indices_keypoints = topk.indices  # B x top_k
        else:
            if self.scores_th > 0:
                masks = nms_scores > self.scores_th
                if masks.sum() == 0:
                    th = scores_nograd.reshape(b, -1).mean(dim=1)  # th = self.scores_th
                    masks = nms_scores > th.reshape(b, 1, 1, 1)
            else:
                th = scores_nograd.reshape(b, -1).mean(dim=1)  # th = self.scores_th
                masks = nms_scores > th.reshape(b, 1, 1, 1)
            masks = masks.reshape(b, -1)

            indices_keypoints = []  # list, B x (any size)
            scores_view = scores_nograd.reshape(b, -1)
            for mask, scores in zip(masks, scores_view):
                indices = mask.nonzero(as_tuple=False)[:, 0]
                if len(indices) > self.n_limit:
                    kpts_sc = scores[indices]
                    sort_idx = kpts_sc.sort(descending=True)[1]
                    sel_idx = sort_idx[:self.n_limit]
                    indices = indices[sel_idx]
                indices_keypoints.append(indices)

        keypoints = []
        scoredispersitys = []
        kptscores = []
        if sub_pixel:
            # detect soft keypoints with grad backpropagation
            patches = self.unfold(scores_map)  # B x (kernel**2) x (H*W)
            self.hw_grid = self.hw_grid.to(patches)  # to device
            for b_idx in range(b):
                patch = patches[b_idx].t()  # (H*W) x (kernel**2)
                indices_kpt = indices_keypoints[b_idx]  # one dimension vector, say its size is M
                patch_scores = patch[indices_kpt]  # M x (kernel**2)

                # max is detached to prevent undesired backprop loops in the graph
                max_v = patch_scores.max(dim=1).values.detach()[:, None]
                x_exp = ((patch_scores - max_v) / self.temperature).exp()  # M * (kernel**2), in [0, 1]

                # \frac{ \sum{(i,j) \times \exp(x/T)} }{ \sum{\exp(x/T)} }
                xy_residual = x_exp @ self.hw_grid / x_exp.sum(dim=1)[:, None]  # Soft-argmax, Mx2

                hw_grid_dist2 = torch.norm((self.hw_grid[None, :, :] - xy_residual[:, None, :]) / self.radius,
                                           dim=-1) ** 2
                scoredispersity = (x_exp * hw_grid_dist2).sum(dim=1) / x_exp.sum(dim=1)

                # compute result keypoints
                keypoints_xy_nms = torch.stack([indices_kpt % w, indices_kpt // w], dim=1)  # Mx2
                keypoints_xy = keypoints_xy_nms + xy_residual
                keypoints_xy = keypoints_xy / keypoints_xy.new_tensor(
                    [w - 1, h - 1]) * 2 - 1  # (w,h) -> (-1~1,-1~1)

                kptscore = torch.nn.functional.grid_sample(scores_map[b_idx].unsqueeze(0),
                                                           keypoints_xy.view(1, 1, -1, 2),
                                                           mode='bilinear', align_corners=True)[0, 0, 0, :]  # CxN

                keypoints.append(keypoints_xy)
                scoredispersitys.append(scoredispersity)
                kptscores.append(kptscore)
        else:
            for b_idx in range(b):
                indices_kpt = indices_keypoints[b_idx]  # one dimension vector, say its size is M
                keypoints_xy_nms = torch.stack([indices_kpt % w, indices_kpt // w], dim=1)  # Mx2
                keypoints_xy = keypoints_xy_nms / keypoints_xy_nms.new_tensor(
                    [w - 1, h - 1]) * 2 - 1  # (w,h) -> (-1~1,-1~1)
                kptscore = torch.nn.functional.grid_sample(scores_map[b_idx].unsqueeze(0),
                                                           keypoints_xy.view(1, 1, -1, 2),
                                                           mode='bilinear', align_corners=True)[0, 0, 0, :]  # CxN
                keypoints.append(keypoints_xy)
                scoredispersitys.append(None)
                kptscores.append(kptscore)

        return keypoints, scoredispersitys, kptscores

    def forward(self, scores_map, descriptor_map, sub_pixel=False):
        """
        :param scores_map:  Bx1xHxW
        :param descriptor_map: BxCxHxW
        :param sub_pixel: whether to use sub-pixel keypoint detection
        :return: kpts: list[Nx2,...]; kptscores: list[N,....] normalised position: -1.0 ~ 1.0
        """
        keypoints, scoredispersitys, kptscores = self.detect_keypoints(scores_map,
                                                                       sub_pixel)

        descriptors = sample_descriptor(descriptor_map, keypoints, sub_pixel)

        # keypoints: B M 2
        # descriptors: B M D
        # scoredispersitys:
        return keypoints, descriptors, kptscores, scoredispersitys


import torch
import numpy as np
import scipy

from easy_local_features.submodules.git_d2net.models import D2Net, preprocess_image, process_multiscale
from ..utils import ops
from .basemodel import BaseExtractor, MethodType
from ..utils.download import downloadModel
from ..matching.nearest_neighbor import NearestNeighborMatcher

from omegaconf import OmegaConf
from typing import TypedDict


class D2NetConfig(TypedDict):
    model_name: str
    top_k: int
    detection_threshold: float
    nms_radius: int
    use_relu: bool


class D2Net_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE

    default_conf: D2NetConfig = {
        "model_name": "d2net",
        "top_k": 2048,
        "detection_threshold": 0.2,
        "nms_radius": 4,
        "use_relu": True,
    }

    def __init__(self, conf: D2NetConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        top_kps = conf.top_k
        use_relu = conf.use_relu

        self.DEV = torch.device("cpu")
        self.model = D2Net(use_relu=use_relu)

        model_file_url = (
            "https://dusmanu.com/files/d2-net/d2_tf_no_phototourism.pth"  # https://dsmn.ml/files/d2-net/d2_ots.pth
        )
        model_path = downloadModel("d2net", "d2_tf_no_phototourism.pth", model_file_url)
        self.model.load_state_dict(torch.load(model_path, map_location=self.DEV)["model"])

        self.model.eval()
        self.model.to(self.DEV)
        self.top_kps = top_kps

        self.max_edge = 1600
        self.max_sum_edges = 2800
        self.multiscale = False
        self.use_relu = use_relu
        self.preprocessing = "torch"  # 'caffe' or 'torch'

        self.matcher = NearestNeighborMatcher()

    def compute(self, img, cv_kps):
        raise NotImplementedError

    def detect(self, img, op=None):
        raise NotImplementedError

    def detectAndCompute(self, image, return_dict=False):
        image = ops.prepareImage(image)
        image = ops.to_cv(image)

        if len(image.shape) == 2:
            image = image[:, :, np.newaxis]
            image = np.repeat(image, 3, -1)

        # TODO: switch to PIL.Image due to deprecation of scipy.misc.imresize.
        resized_image = image
        if max(resized_image.shape) > self.max_edge:
            resized_image = scipy.misc.imresize(resized_image, self.max_edge / max(resized_image.shape)).astype("float")
        if sum(resized_image.shape[:2]) > self.max_sum_edges:
            resized_image = scipy.misc.imresize(
                resized_image, self.max_sum_edges / sum(resized_image.shape[:2])
            ).astype("float")

        fact_i = image.shape[0] / resized_image.shape[0]
        fact_j = image.shape[1] / resized_image.shape[1]

        input_image = preprocess_image(resized_image, preprocessing=self.preprocessing)

        with torch.no_grad():
            if self.multiscale:
                keypoints, scores, descriptors = process_multiscale(
                    torch.tensor(input_image[np.newaxis, :, :, :].astype(np.float32), device=self.DEV), self.model
                )
            else:
                keypoints, scores, descriptors = process_multiscale(
                    torch.tensor(input_image[np.newaxis, :, :, :].astype(np.float32), device=self.DEV),
                    self.model,
                    scales=[1],
                )

        # Input image coordinates
        keypoints[:, 0] *= fact_i
        keypoints[:, 1] *= fact_j
        # i, j -> u, v
        keypoints = keypoints[:, [1, 0, 2]]
        keypoints, scales = keypoints[:, :2], keypoints[:, 2]

        argsort = np.argsort(-scores)

        keypoints = keypoints[argsort[: self.top_kps]].unsqueeze(0)
        scores = scores[argsort[: self.top_kps]].unsqueeze(0)
        descriptors = descriptors[argsort[: self.top_kps]].unsqueeze(0)
        scales = scales[argsort[: self.top_kps]].unsqueeze(0)

        if return_dict:
            return {"keypoints": keypoints, "descriptors": descriptors, "scores": scores, "scales": scales}

        return keypoints, descriptors

    def detect(self, img):
        keypoints, descriptors = self.detectAndCompute(img)
        return keypoints

    def compute(self, img, keypoints):
        raise NotImplementedError

    def to(self, device):
        self.model.to(device)
        self.DEV = device

    @property
    def has_detector(self):
        return True



import numpy as np
import torch
import torchvision
import cv2
import functools
from pathlib import Path

def fromPath(path, gray=False, batch=True, imagenet=False):
    if isinstance(path, Path):
        path = str(path)
        
    if path.endswith(".ppm"):
        import cv2
        im = cv2.imread(path)
        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
        im = torch.from_numpy(im).float() / 255.0
        im = im.permute(2, 0, 1)
    else:
        im = torchvision.io.read_image(path).float() / 255.0

    if imagenet:
        im = torchvision.transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225]
        )(im)
    if gray:
        im = im.mean(0, keepdim=True)
    if batch:
        im = im.unsqueeze(0)
    return im

def writeKeypoints(keypoints, filename):
    """
    Write keypoints to a file using numpy compressed format
    """

    total_keypoints = len(keypoints)
    np_keypoints = np.zeros((total_keypoints, 4))
    for i, kp in enumerate(keypoints):
        np_keypoints[i, :] = [kp.pt[0], kp.pt[1], kp.size, kp.angle]

    np.savez_compressed(filename, keypoints=np_keypoints)

@functools.lru_cache(maxsize=128)
def readKeypoints(filename):
    """
    Read keypoints from a file
    """
    
    npzfile = np.load(filename)
    keypoints = []
    for row in npzfile['keypoints']:
        kp = cv2.KeyPoint(x=row[0], y=row[1], size=row[2], angle=row[3])
        keypoints.append(kp)

    return keypoints

def writeDescriptors(descriptors, filename):
    """
    Write descriptors to a file using numpy compressed format
    """

    np.savez_compressed(filename, descriptors=descriptors)

@functools.lru_cache(maxsize=128)
def readDescriptors(filename):
    """
    Read descriptors from a file
    """
    
    npzfile = np.load(filename)
    return npzfile['descriptors']



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

# References:
#   https://github.com/facebookresearch/dino/blob/master/vision_transformer.py
#   https://github.com/rwightman/pytorch-image-models/tree/master/timm/layers/patch_embed.py

from typing import Callable, Optional, Tuple, Union

from torch import Tensor
import torch.nn as nn


def make_2tuple(x):
    if isinstance(x, tuple):
        assert len(x) == 2
        return x

    assert isinstance(x, int)
    return (x, x)


class PatchEmbed(nn.Module):
    """
    2D image to patch embedding: (B,C,H,W) -> (B,N,D)

    Args:
        img_size: Image size.
        patch_size: Patch token size.
        in_chans: Number of input image channels.
        embed_dim: Number of linear projection output channels.
        norm_layer: Normalization layer.
    """

    def __init__(
        self,
        img_size: Union[int, Tuple[int, int]] = 224,
        patch_size: Union[int, Tuple[int, int]] = 16,
        in_chans: int = 3,
        embed_dim: int = 768,
        norm_layer: Optional[Callable] = None,
        flatten_embedding: bool = True,
    ) -> None:
        super().__init__()

        image_HW = make_2tuple(img_size)
        patch_HW = make_2tuple(patch_size)
        patch_grid_size = (
            image_HW[0] // patch_HW[0],
            image_HW[1] // patch_HW[1],
        )

        self.img_size = image_HW
        self.patch_size = patch_HW
        self.patches_resolution = patch_grid_size
        self.num_patches = patch_grid_size[0] * patch_grid_size[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        self.flatten_embedding = flatten_embedding

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_HW, stride=patch_HW)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        _, _, H, W = x.shape
        patch_H, patch_W = self.patch_size

        assert H % patch_H == 0, f"Input image height {H} is not a multiple of patch height {patch_H}"
        assert W % patch_W == 0, f"Input image width {W} is not a multiple of patch width: {patch_W}"

        x = self.proj(x)  # B C H W
        H, W = x.size(2), x.size(3)
        x = x.flatten(2).transpose(1, 2)  # B HW C
        x = self.norm(x)
        if not self.flatten_embedding:
            x = x.reshape(-1, H, W, self.embed_dim)  # B H W C
        return x

    def flops(self) -> float:
        Ho, Wo = self.patches_resolution
        flops = Ho * Wo * self.embed_dim * self.in_chans * (self.patch_size[0] * self.patch_size[1])
        if self.norm is not None:
            flops += Ho * Wo * self.embed_dim
        return flops



from .utils import (
    pose_auc,
    get_pose,
    compute_relative_pose,
    compute_pose_error,
    estimate_pose,
    estimate_pose_uncalibrated,
    rotate_intrinsic,
    get_tuple_transform_ops,
    get_depth_tuple_transform_ops,
    warp_kpts,
    numpy_to_pil,
    tensor_to_pil,
    recover_pose,
    signed_left_to_right_epipolar_distance,
)
